# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

import uuid
from collections import defaultdict
from collections.abc import Iterable
from functools import cached_property
from pydoc import locate
from typing import Annotated, Any, Literal

from compiler_config.config import InlineResultsProcessing
from pydantic import (
    BeforeValidator,
    ConfigDict,
    Field,
    computed_field,
    field_serializer,
    field_validator,
)
from typing_extensions import TypeAliasType

from qat.purr.compiler.instructions import (
    BinaryOperator as LegacyBinaryOperator,
    IndexAccessor as LegacyIndexAccessor,
    Variable as LegacyVariable,
)
from qat.purr.utils.logger import get_default_logger
from qat.utils.pydantic import (
    AllowExtraFieldsModel,
    FrozenSet,
    NoExtraFieldsModel,
    QubitId,
    ValidatedList,
    ValidatedSet,
    _validate_set,
)

log = get_default_logger()


### Instructions

BASE_INSTR = "qat.ir.instructions.Instruction"


class Instruction(AllowExtraFieldsModel):
    @computed_field
    @cached_property
    def instr_type(self) -> str:
        """
        Returns the type of the instruction, which is the class name.
        """
        return self.__class__.__module__ + "." + self.__class__.__name__

    def __iter__(self):
        yield self

    def __reversed__(self):
        yield self

    def __repr__(self):
        return f"{self.__class__.__name__}()"

    @property
    def number_of_instructions(self):
        return 1

    @property
    def head(self):
        return self

    @property
    def tail(self):
        return self


class InstructionBlock(Instruction, Iterable):
    instructions: ValidatedList[Instruction] = []

    def add(self, *instructions: Instruction, flatten: bool = False):
        for instruction in instructions:
            if flatten:
                for sub_instruction in instruction:
                    self.instructions.append(sub_instruction)
            else:
                self.instructions.append(instruction)
        return self

    def __iter__(self):
        """Iterator over the flattened instructions in the block."""
        for instruction in self.instructions:
            yield from instruction

    def __reversed__(self):
        """Reversed iterator over the flattened instructions in the block."""
        for instruction in reversed(self.instructions):
            yield from reversed(instruction)

    @property
    def number_of_instructions(self):
        return sum(1 for _ in self)

    @property
    def head(self):
        if self.number_of_instructions:
            return self.instructions[0]
        raise IndexError(
            f"Cannot access first element of an empty `{self.__class__.__name__}`."
        )

    @property
    def tail(self):
        if self.number_of_instructions:
            return self.instructions[-1]
        raise IndexError(
            f"Cannot access last element of an empty `{self.__class__.__name__}`."
        )

    def flatten(self):
        flattened_instructions = []
        for instr in self:
            flattened_instructions.append(instr)
        self.instructions = flattened_instructions
        return self

    @field_serializer("instructions")
    def _serialize_instructions(self, instructions: list[Instruction]):
        return [inst.model_dump() for inst in self]

    @field_validator("instructions", mode="before")
    @classmethod
    def _rehydrate_instructions(cls, instructions):
        if isinstance(instructions, Instruction):
            return instructions
        if not isinstance(instructions, (list, ValidatedList)):
            raise TypeError(
                f"Expected `instructions` to be a list of `Instruction` instances or dictionaries, got {type(instructions)}."
            )

        rehydrated = []
        # Cache for located types to avoid repeated lookups
        type_cache: dict[str, type[Instruction]] = {}

        for instr in instructions:
            if isinstance(instr, dict):
                instr_type_str = instr.get("instr_type", BASE_INSTR)

                if instr_type_str not in type_cache:
                    type_cache[instr_type_str] = locate(instr_type_str)
                instr_type = type_cache[instr_type_str]
                rehydrated.append(instr_type(**instr))
            elif isinstance(instr, Instruction):
                rehydrated.append(instr)
            else:
                raise TypeError(
                    f"Instruction must be an Instruction instance or dict, got {type(instr)}."
                )

        return rehydrated


class Repeat(Instruction):
    """
    Global meta-instruction that applies to the entire list of instructions. Repeat
    value of the current operations, also known as shots.
    """

    repeat_count: int | None = None


class EndRepeat(Instruction): ...


class Return(Instruction):
    """A statement defining what to return from a quantum execution."""

    variables: list[str] = []

    @field_validator("variables", mode="before")
    @classmethod
    def _variables_as_list(cls, variables):
        variables = (
            []
            if variables is None
            else ([variables] if not isinstance(variables, list) else variables)
        )
        return variables


class ResultsProcessing(Instruction):
    """
    A meta-instruction that stores how the results are processed.
    """

    variable: str
    results_processing: InlineResultsProcessing = InlineResultsProcessing.Raw

    @classmethod
    def _from_legacy(cls, legacy_rp):
        # private as we dont want to support this in the long-term
        return cls(
            variable=legacy_rp.variable, results_processing=legacy_rp.results_processing
        )


### Variables
class Variable(Instruction):
    """
    States that this value is actually a variable that should be fetched instead.
    """

    name: str
    var_type: type | None = None
    value: Any = None

    @staticmethod
    def with_random_name():
        return Variable(name=str(uuid.uuid4()))

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name}, type={self.var_type}, value={self.value})"

    @field_serializer("var_type", when_used="json")
    def _serialise_type(self, var_type: type) -> str:
        # Types can't be (de)serialized, so we serialise as a string.
        return var_type.__name__

    @field_validator("var_type", mode="before")
    @classmethod
    def _validate_var_type(cls, var_type):
        # Types can't be (de)serialised, so we use a validator to manually
        # find the correct type.
        if isinstance(var_type, str):
            var_type = locate(var_type)
        return var_type

    @field_validator("value", mode="after")
    @classmethod
    def _validate_value_type(cls, value, val_info):
        var_type = val_info.data["var_type"]
        if var_type is not None and not isinstance(value, var_type) and value is not None:
            raise ValueError(f"Value provided has type {type(value)}: must be {var_type}")
        return value


class Label(Instruction):
    """
    Label to apply to a line of code. Used as anchors for other instructions like jumps.
    """

    name: str

    @staticmethod
    def with_random_name():
        return Label(name=f"label_{uuid.uuid4()}")

    def __repr__(self):
        return f"{self.__class__.__name__}(name={self.name})"


class Jump(Instruction):
    """
    Classic jump instruction, should be linked to label with an optional condition.
    """

    label: Label | str
    condition: BinaryOperator | None = None

    @field_validator("label", mode="after")
    @classmethod
    def _validate_label(cls, label: str | Label):
        if isinstance(label, str):
            label = Label(name=label)
        return label

    @property
    def target(self):
        return self.label.name

    def __repr__(self):
        if self.condition is not None:
            return f"Jump if {str(self.condition)} -> {str(self.target)}"
        else:
            return f"Jump -> {str(self.target)}"


class LoopCount(int): ...


### Binary Operators


class BinaryOperator(NoExtraFieldsModel):
    """Binary operator, such as ``x == y``, ``x != y`` etc."""

    left: int | float | Variable
    right: int | float | Variable


class Equals(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} == {str(self.right)}"


class NotEquals(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} != {str(self.right)}"


class GreaterThan(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} > {str(self.right)}"


class GreaterOrEqualThan(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} >= {str(self.right)}"


class LessThan(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} < {str(self.right)}"


class LessOrEqualThan(BinaryOperator):
    def __repr__(self):
        return f"{str(self.left)} <= {str(self.right)}"


class Plus(BinaryOperator):
    # TODO: Improve operators in Pydantic version
    def __repr__(self):
        return f"{str(self.left)} + {str(self.right)}"


AssignTypes = int | float | str | Variable | BinaryOperator
RecursiveAssignTypes = TypeAliasType(
    "ResursiveAssignTypes", "AssignTypes | list[RecursiveAssignTypes]"
)


class Assign(Instruction):
    """
    Assigns a value to a variable.

    This is used to assign some value (e.g. the results from an acquisition) to a variable.
    In the legacy instructions, `Assign` could be given a `Variable` that declares the value
    as another variable. It could also be given more complex structures, should as a list of
    `Variables`, or an `IndexAccessor`. For example, this could be used to assign each of the
    qubit acquisitions to a single register,

    ```
        c = [
            Qubit 1 measurements,
            Qubit 2 measurements,
            ...
        ]
    ```

    In general, declarations and allocations could be improved in future iterations, and
    functionality may change with improvements to the front end. For now, `Assign` has been
    adapted to support the required front end behaviour. The value is allowed to be a list
    (with recurssions supported) of strings that declare what Variable to point to, or a tuple
    for `IndexAccessor`.
    """

    name: str
    value: RecursiveAssignTypes

    @classmethod
    def _from_legacy(cls, legacy_assign):
        # private as we dont want to support this in the long-term

        def recursively_strip(value):
            if isinstance(value, list):
                return [recursively_strip(val) for val in value]
            elif isinstance(value, LegacyIndexAccessor):
                return (value.name, value.index)
            elif isinstance(value, LegacyVariable):
                return value.name
            elif isinstance(value, LegacyBinaryOperator):
                return str(value).replace("variable ", "")
            return value

        return cls(name=legacy_assign.name, value=recursively_strip(legacy_assign.value))

    def __repr__(self):
        return f"Assign {self.name} = {str(self.value)}"


### Quantum Instructions
class QuantumInstruction(Instruction):
    """
    Any node that deals particularly with quantum operations. All quantum operations
    must have some sort of target on the quantum computer, such as a qubit, channel, or
    another form of component.
    """

    targets: Annotated[FrozenSet[str], BeforeValidator(_validate_set)] | None = Field(
        min_length=1, max_length=1, alias="target"
    )
    duration: float = Field(ge=0, default=0)  # in seconds

    model_config = ConfigDict(validate_by_name=True, populate_by_name=True)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(duration={self.duration}, targets={self.targets})"
        )

    @property
    def target(self):
        return next(iter(self.targets))


class QuantumInstructionBlock(QuantumInstruction, InstructionBlock):
    instructions: ValidatedList[QuantumInstruction] = []
    targets: Annotated[ValidatedSet[str], BeforeValidator(_validate_set)] = Field(
        default_factory=lambda: ValidatedSet[str]()
    )
    _duration_per_target: dict[str, float] = defaultdict(lambda: 0.0, dict())

    def add(self, *instructions: QuantumInstruction):
        super().add(*instructions)
        for instruction in instructions:
            for target in instruction.targets:
                self.targets.add(target)
                self._duration_per_target[target] += instruction.duration

            self.duration = max(self._duration_per_target.values())

        return self


class PhaseShift(QuantumInstruction):
    """
    A PhaseShift instruction is used to change the phase of waveforms sent down
    the pulse channel.
    """

    phase: float = 0.0


class FrequencySet(QuantumInstruction):
    """Set the frequency of a pulse channel."""

    frequency: float = 0.0


class FrequencyShift(QuantumInstruction):
    """Change the frequency of a pulse channel."""

    frequency: float = 0.0


class Delay(QuantumInstruction):
    """Instructs a quantum target to do nothing for a fixed time."""


class Synchronize(QuantumInstruction):
    """
    Tells the QPU to wait for all the target channels to be free before continuing
    execution on any of them.
    """

    duration: Literal[0] = 0

    targets: Annotated[FrozenSet[str], BeforeValidator(_validate_set)] = Field(min_length=2)


class PhaseSet(QuantumInstruction):
    """Sets the phase for a pulse channel.

    This sets the absolute phase of an NCO, unlike the :class:`PhaseShift`, which changes
    the phase relative to the current phase.
    """

    phase: float = 0.0


class PhaseReset(PhaseSet):
    phase: Literal[0.0] = 0.0


class Reset(QuantumInstruction):
    """Resets this qubit to its starting state."""

    targets: Annotated[FrozenSet[str], BeforeValidator(_validate_set)] = Field(
        max_length=0, default=FrozenSet[str](set())
    )
    qubit_targets: Annotated[FrozenSet[QubitId], BeforeValidator(_validate_set)] = Field(
        min_length=1, max_length=1, alias="qubit_target"
    )

    @property
    def target(self):
        return None

    @property
    def qubit_target(self):
        return next(iter(self.qubit_targets))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

import itertools
from collections import OrderedDict, defaultdict
from copy import copy, deepcopy
from numbers import Number

import numpy as np
from compiler_config.config import CompilerConfig, MetricsType
from more_itertools import partition

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import TransformPass
from qat.core.result_base import ResultManager
from qat.middleend.passes.purr.analysis import ActiveChannelResults
from qat.model.target_data import TargetData
from qat.purr.backends.qiskit_simulator import QiskitBuilder, QiskitBuilderWrapper
from qat.purr.backends.utilities import evaluate_shape
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, Qubit
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Acquire,
    AcquireMode,
    Assign,
    CustomPulse,
    Delay,
    DeviceUpdate,
    EndRepeat,
    EndSweep,
    GreaterThan,
    Instruction,
    Jump,
    Label,
    MeasurePulse,
    PhaseReset,
    PhaseSet,
    PhaseShift,
    Plus,
    PostProcessing,
    PostProcessType,
    ProcessAxis,
    Pulse,
    PulseShapeType,
    QuantumInstruction,
    Repeat,
    Reset,
    Return,
    Sweep,
    Synchronize,
    Variable,
)
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class IntegratorAcquireSanitisation(TransformPass):
    """Changes `AcquireMode.INTEGRATOR` acquisitions to `AcquireMode.RAW`.

    The legacy echo/RTCS engines expect the acquisition mode to be either `RAW` or `SCOPE`.
    While the actual execution can process `INTEGRATOR` by treating it as `RAW`, they are
    typically santitised the runtime using :meth:`EchoEngine.optimize()`. If not done in the
    new pipelines, it will conflict with :class:`PostProcessingSantisiation`, and return the
    wrong results. The new echo engine supports all acquisition modes, so this is not a
    problem here.
    """

    def run(
        self,
        ir: InstructionBuilder,
        *args,
        **kwargs,
    ):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""
        for inst in [instr for instr in ir.instructions if isinstance(instr, Acquire)]:
            if inst.mode == AcquireMode.INTEGRATOR:
                inst.mode = AcquireMode.RAW
        return ir


class QiskitInstructionsWrapper(TransformPass):
    """Wraps the Qiskit builder in a wrapper to match the pipelines API.

    A really silly pass needed to wrap the :class:`QiskitBuilder` in an object that allows
    `QiskitBuilderWrapper.instructions` to be called, allowing the builder to be used in the
    the :class:`LegacyRuntime`. This is needed because the qiskit engine has a different API
    to other `purr` engines, requiring the whole builder to be passed (as opposed to
    `builder.instructions`).
    """

    def run(self, ir: QiskitBuilder, *args, **kwargs) -> QiskitBuilderWrapper:
        """:param ir: The Qiskit instructions"""
        return QiskitBuilderWrapper(ir)


class LegacyPhaseOptimisation(TransformPass):
    """Iterates through the list of instructions and compresses contiguous
    :class:`PhaseShift` instructions.
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        :param met_mgr: The metrics manager to store the number of instructions after
            optimisation.
        """

        accum_phaseshifts: dict[PulseChannel, PhaseShift] = {}
        optimized_instructions: list[Instruction] = []
        for instruction in ir.instructions:
            if isinstance(instruction, PhaseShift) and isinstance(
                instruction.phase, Number
            ):
                if accum_phaseshift := accum_phaseshifts.get(instruction.channel, None):
                    accum_phaseshift.phase += instruction.phase
                else:
                    accum_phaseshifts[instruction.channel] = PhaseShift(
                        instruction.channel, instruction.phase
                    )
            elif isinstance(instruction, (Pulse, CustomPulse)):
                quantum_targets = getattr(instruction, "quantum_targets", [])
                if not isinstance(quantum_targets, list):
                    quantum_targets = [quantum_targets]
                for quantum_target in quantum_targets:
                    if quantum_target in accum_phaseshifts:
                        optimized_instructions.append(accum_phaseshifts.pop(quantum_target))
                optimized_instructions.append(instruction)
            elif isinstance(instruction, PhaseReset):
                for channel in instruction.quantum_targets:
                    accum_phaseshifts.pop(channel, None)
                optimized_instructions.append(instruction)
            else:
                optimized_instructions.append(instruction)

        ir.instructions = optimized_instructions
        met_mgr.record_metric(
            MetricsType.OptimizedInstructionCount, len(optimized_instructions)
        )
        return ir


class LoopCount(int): ...


class RepeatTranslation(TransformPass):
    """Transform :class:`Repeat` instructions so that they are replaced with:
    :class:`Variable`, :class:`Assign`, and :class:`Label` instructions at the start,
    and :class:`Assign` and :class:`Jump` instructions at the end."""

    def __init__(self, target_data: TargetData):
        self.target_data = target_data

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> InstructionBuilder:
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""
        instructions: list[Instruction] = []
        label_data: OrderedDict[Label, dict[str, Repeat | Variable]] = OrderedDict()
        repetition_period = None
        passive_reset_time = None

        def _close_repeat(label: Label, data: dict, index: int = None) -> None:
            # TODO: Adjust so that closing is done before finishing instructions such as
            # returns and postprocessing. COMPILER-452
            repeat = data["repeat"]
            var = data["var"]
            assign = Assign(var.name, Plus(var, 1))
            jump = Jump(label, GreaterThan(repeat.repeat_count, var))
            instructions.extend([assign, jump])

        for inst in ir.instructions:
            if isinstance(inst, Repeat):
                label = Label.with_random_name(ir.existing_names)
                var = Variable(label.name + "_count", LoopCount)
                assign = Assign(var.name, 0)
                label_data[label] = {"repeat": inst, "var": var}
                instructions.extend([var, assign, label])
                if repetition_period is None and inst.repetition_period is not None:
                    repetition_period = inst.repetition_period
                if passive_reset_time is None and inst.passive_reset_time is not None:
                    passive_reset_time = inst.passive_reset_time
            elif isinstance(inst, EndRepeat):
                if len(label_data) < 1:
                    raise ValueError("EndRepeat found without associated Repeat.")
                label, data = label_data.popitem()
                _close_repeat(label, data)
            else:
                instructions.append(inst)

        for label, data in reversed(label_data.items()):
            _close_repeat(label, data)

        if passive_reset_time is None and repetition_period is None:
            passive_reset_time = self.target_data.QUBIT_DATA.passive_reset_time
        ir.repetition_period = repetition_period
        ir.passive_reset_time = passive_reset_time

        ir.instructions = instructions
        return ir


class PhaseOptimisation(TransformPass):
    """Iterates through the list of instructions and compresses contiguous
    :class:`PhaseShift` instructions. This pass will change :class:`PhaseReset` to
    :class:`PhaseSet` instructions.
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        :param met_mgr: The metrics manager to store the number of instructions after
            optimisation.
        """

        accum_phaseshifts: dict[str, PhaseShift | PhaseSet] = dict()
        optimized_instructions: list[Instruction] = []
        for instruction in ir.instructions:
            if isinstance(instruction, (PhaseShift, PhaseSet)) and isinstance(
                instruction.phase, Number
            ):
                key = instruction.channel.partial_id()
                accum_phaseshifts[key] = self.merge_phase_instructions(
                    instruction.channel, accum_phaseshifts.get(key, None), instruction
                )
            elif isinstance(instruction, PhaseReset):
                for channel in instruction.quantum_targets:
                    key = channel.partial_id()
                    accum_phaseshifts[key] = self.merge_phase_instructions(
                        channel, accum_phaseshifts.get(key, None), instruction
                    )
            elif isinstance(
                instruction, (CustomPulse, Pulse, Acquire, PhaseSet, PhaseShift)
            ):
                quantum_targets = getattr(instruction, "quantum_targets", [])
                if not isinstance(quantum_targets, list):
                    quantum_targets = [quantum_targets]
                for quantum_target in quantum_targets:
                    if quantum_target.partial_id() in accum_phaseshifts:
                        new_instruction = accum_phaseshifts.pop(quantum_target.partial_id())
                        if not (
                            isinstance(new_instruction, PhaseShift)
                            and np.isclose(new_instruction.phase % (2 * np.pi), 0.0)
                        ):
                            optimized_instructions.append(new_instruction)
                optimized_instructions.append(instruction)
            elif isinstance(instruction, (Delay, Synchronize)):
                for channel in instruction.quantum_targets:
                    key = channel.partial_id()
                    if isinstance(accum_phaseshifts.get(key, None), PhaseSet):
                        optimized_instructions.append(accum_phaseshifts.pop(key))
                optimized_instructions.append(instruction)
            elif isinstance(instruction, (Jump, Label)):
                accum_phaseshifts = dict()
                optimized_instructions.append(instruction)
            else:
                optimized_instructions.append(instruction)

        ir.instructions = optimized_instructions
        met_mgr.record_metric(
            MetricsType.OptimizedInstructionCount, len(optimized_instructions)
        )
        return ir

    @staticmethod
    def merge_phase_instructions(
        target: PulseChannel,
        phase1: PhaseSet | PhaseReset | PhaseShift | None,
        phase2: PhaseSet | PhaseReset | PhaseShift,
    ):
        if isinstance(phase2, PhaseReset):
            return PhaseSet(target, 0.0)
        elif isinstance(phase2, PhaseSet):
            return PhaseSet(target, phase2.phase)
        elif phase1 is None:
            return phase2
        elif isinstance(phase1, PhaseReset):
            return PhaseSet(target, phase2.phase)
        elif isinstance(phase1, PhaseSet):
            return PhaseSet(target, phase1.phase + phase2.phase)
        else:
            return PhaseShift(target, phase1.phase + phase2.phase)


class PostProcessingSanitisation(TransformPass):
    """Checks that the :class:`PostProcessing` instructions that follow an acquisition are
    suitable for the acquisition mode, and removes them if not.

    Extracted from :meth:`qat.purr.backends.live.LiveDeviceEngine.optimize`.
    """

    def run(
        self,
        ir: InstructionBuilder,
        _: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param met_mgr: The metrics manager to store the number of instructions after
            optimisation.
        """

        # TODO -COMPILER-851 - Specify ctrl HW features in TargetData
        enable_hw_averaging = kwargs.get("enable_hw_averaging", False)

        pp_insts = [val for val in ir.instructions if isinstance(val, PostProcessing)]
        discarded = []
        for pp in pp_insts:
            if pp.acquire.mode == AcquireMode.SCOPE:
                if (
                    not enable_hw_averaging
                    and pp.process == PostProcessType.MEAN
                    and ProcessAxis.SEQUENCE in pp.axes
                    and len(pp.axes) <= 1
                ):
                    discarded.append(pp)

            elif pp.acquire.mode == AcquireMode.INTEGRATOR:
                if (
                    pp.process == PostProcessType.DOWN_CONVERT
                    and ProcessAxis.TIME in pp.axes
                    and len(pp.axes) <= 1
                ):
                    discarded.append(pp)
                if (
                    pp.process == PostProcessType.MEAN
                    and ProcessAxis.TIME in pp.axes
                    and len(pp.axes) <= 1
                ):
                    discarded.append(pp)
        ir.instructions = [val for val in ir.instructions if val not in discarded]
        met_mgr.record_metric(MetricsType.OptimizedInstructionCount, len(ir.instructions))
        return ir


class AcquireSanitisation(TransformPass):
    """Sanitises the :class:`Acquire` instruction: the first per pulse channel is split into
    an :class:`Acquire` and a :class:`Delay`, and other acquisitions have their delay
    removed.

    :class:`Acquire` instructions are defined by a "duration" for which they instruct the
    target to readout. They also contain a "delay" attribute, which instructions the
    acquisition to start after some given time. This pass separates acqusitions with a
    delay into two instructions for the first acquire that acts on the channel. For multiple
    acquisitions on a single channel, the delay is not needed for the following
    acquisitions, and is set to zero.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        """

        new_instructions: list[Instruction] = []

        for inst in ir.instructions:
            if isinstance(inst, Acquire):
                if inst.delay:
                    delay = Delay(inst.quantum_targets, inst.delay)
                    inst.delay = 0.0
                    new_instructions.extend([delay, inst])
                else:
                    new_instructions.append(inst)
            else:
                new_instructions.append(inst)
        ir.instructions = new_instructions
        return ir


class InstructionGranularitySanitisation(TransformPass):
    """Rounds the durations of quantum instructions so they are multiples of the clock
    cycle.

    Only supports quantum instructions with static non-zero durations. Assumes that
    instructions with a non-zero duration only act on a single pulse channel. The
    santisiation is done for all instructions simultaneously using numpy for performance.

    For :class:`CustomPulse` instructions, the durations are rounded up by padding the pulse
    with zero amplitude at the end. For other relevant instructions, we round down: this is
    for compatibility with calibration files that are calibrated using legacy code. However,
    in the future we might consider changing this to round up for consistency.

    .. warning::

        This pass has the potential to invalidate the timings for sequences of instructions
        that are time-sensitive. For example, if a pulse has an invalid time, it will round
        it up to the nearest integer multiple. Furthemore, it will assume that
        :class:`Acquire` instructions have no delay. This can be forced explicitly using the
        :class:`AcquireSanitisation` pass.
    """

    def __init__(self, model: QuantumHardwareModel, target_data: TargetData):
        """:param target_data: Target-related information."""

        self.clock_cycle = target_data.clock_cycle
        qubit_sample_times = {
            qubit.physical_channel: target_data.QUBIT_DATA.sample_time
            for qubit in model.qubits
        }
        res_sample_times = {
            resonator.physical_channel: target_data.RESONATOR_DATA.sample_time
            for resonator in model.resonators
        }
        self.sample_times = qubit_sample_times | res_sample_times

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        """

        self.sanitise_quantum_instructions(
            [inst for inst in ir.instructions if isinstance(inst, (Pulse, Acquire, Delay))]
        )
        self.sanitise_custom_pulses(
            [inst for inst in ir.instructions if isinstance(inst, CustomPulse)]
        )
        self.sanitise_acquire_filters(
            [inst for inst in ir.instructions if isinstance(inst, Acquire)]
        )
        return ir

    def sanitise_quantum_instructions(self, instructions: list[Pulse | Acquire | Delay]):
        """Sanitises the durations quantum instructions with non-zero duration by rounding
        down to the nearest clock cycle."""

        durations = np.asarray([inst.duration for inst in instructions])
        multiples = durations / self.clock_cycle
        rounded_multiples = np.floor(multiples + 1e-10).astype(int)
        # 1e-10 for floating point errors
        durations_equal = np.isclose(multiples, rounded_multiples)

        invalid_instructions: set[str] = set()
        for idx in np.where(np.logical_not(durations_equal))[0]:
            inst = instructions[idx]
            invalid_instructions.add(str(inst))
            new_duration = rounded_multiples[idx] * self.clock_cycle
            if isinstance(inst, Pulse):
                inst.width = new_duration
            else:
                inst.time = new_duration

        if len(invalid_instructions) >= 1:
            log.info(
                "The following instructions do not have durations that are integer "
                f"multiples of the clock cycle {self.clock_cycle}, and will be rounded "
                "down: " + ", ".join(set(invalid_instructions))
            )

    def sanitise_custom_pulses(self, instructions: list[CustomPulse]):
        """Sanitises the durations of :class:`CustomPulse` instructions by padding the
        pulses with zero amplitudes."""

        durations = np.asarray([inst.duration for inst in instructions])
        multiples = durations / self.clock_cycle
        # 1e-10 for floating point errors
        rounded_multiples = np.ceil(multiples - 1e-10).astype(int)
        durations_equal = np.isclose(multiples, rounded_multiples)

        invalid_instructions: set[str] = set()
        for idx in np.where(np.logical_not(durations_equal))[0]:
            inst = instructions[idx]
            invalid_instructions.add(str(inst))
            new_duration = rounded_multiples[idx] * self.clock_cycle

            padding = int(
                np.round(
                    (new_duration - durations[idx])
                    / self.sample_times[inst.quantum_targets[0].physical_channel],
                    0,
                )
            )

            if isinstance(inst.samples, list):
                inst.samples.extend([0.0 + 0.0j] * padding)
            elif isinstance(inst.samples, np.ndarray):
                inst.samples = np.append(
                    inst.samples, np.zeros(padding, dtype=inst.samples.dtype)
                )
            else:
                raise TypeError(
                    f"Unsupported type for samples in CustomPulse: {type(inst.samples)}"
                )

        if len(invalid_instructions) > 1:
            log.info(
                "The following custom pulses do not have durations that are integer "
                f"multiples of the clock cycle {self.clock_cycle}, and will be rounded "
                "up by padding with zero amplitudes: "
                + ", ".join(set(invalid_instructions))
            )

    def sanitise_acquire_filters(self, instructions: list[Acquire]):
        """Sanitises the durations of :class:`Acquire` filters by matching it to the
        duration of the :class:`Acquire`. For :class:`CustomPulse` filters, this strips
        away the samples that are not needed."""
        for instruction in instructions:
            if isinstance(instruction.filter, Pulse):
                new_filter = copy(instruction.filter)
                new_filter.width = instruction.duration
                instruction.filter = new_filter
            elif isinstance(instruction.filter, CustomPulse):
                num_samples = int(
                    np.round(
                        instruction.duration
                        / self.sample_times[instruction.quantum_targets[0].physical_channel]
                    )
                )
                if instruction.filter.duration > instruction.duration:
                    # if the filter is longer than the acquire, truncate it
                    samples = instruction.filter.samples[:num_samples]
                else:
                    # if the filter is shorter than the acquire, pad it with zeros
                    samples = np.zeros(num_samples, dtype=np.complex128)
                    samples[: len(instruction.filter.samples)] = instruction.filter.samples
                instruction.filter = CustomPulse(
                    instruction.filter.channel,
                    samples,
                    instruction.filter.ignore_channel_scale,
                )


class InitialPhaseResetSanitisation(TransformPass):
    """
    Checks if every active pulse channel has a phase reset in the beginning.

    .. warning::

        This pass implies that an `ActiveChannelAnalysis` is performed prior to this pass.
    """

    def run(self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs):
        active_targets = list(res_mgr.lookup_by_type(ActiveChannelResults).targets)

        if active_targets:
            index = next(
                filter(
                    lambda tup: True if isinstance(tup[1], QuantumInstruction) else None,
                    enumerate(ir._instructions),
                )
            )[0]
            ir.insert(PhaseReset(active_targets), index)

        return ir


class MeasurePhaseResetSanitisation(TransformPass):
    """
    Adds a phase reset before every measure pulse.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        new_instructions = []
        for instr in ir.instructions:
            if isinstance(instr, MeasurePulse):
                new_instructions.append(PhaseReset(instr.quantum_targets))
            new_instructions.append(instr)

        ir.instructions = new_instructions

        return ir


class InactivePulseChannelSanitisation(TransformPass):
    """Removes instructions that act on inactive pulse channels.

    Many channels that aren't actually needed to execute the program contain instructions,
    mainly :class:`Synchronize` instructions and :class:`PhaseShift` instructions which
    can happen when either of the instructions are applied to qubits. To simplify analysis
    and optimisations in later passes, it makes sense to filter these out to reduce the
    overall instruction amount.

    .. note::

        This pass requires results from the
        :class:`ActivePulseChannelAnalysis <qat.middleend.passes.analysis.ActivePulseChannelAnalysis>`
        to be stored in the results manager.
    """

    def run(
        self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs
    ) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        active_channels = [
            target.full_id()
            for target in res_mgr.lookup_by_type(ActiveChannelResults).targets
        ]
        instructions: list[Instruction] = []
        for inst in ir.instructions:
            if isinstance(inst, (PostProcessing, Pulse, CustomPulse, Acquire)):
                # instructions which define an active channel
                instructions.append(inst)
            elif isinstance(inst, Synchronize):
                # inactive channels need stripping from syncs
                inst.quantum_targets = [
                    target
                    for target in inst.quantum_targets
                    if target.full_id() in active_channels
                ]
                if len(inst.quantum_targets) > 0:
                    instructions.append(inst)
            elif isinstance(inst, QuantumInstruction):
                # other instructions need their targets checking
                target = next(iter(inst.quantum_targets))
                if target.full_id() in active_channels:
                    instructions.append(inst)
            else:
                instructions.append(inst)
        ir.instructions = instructions
        return ir


class InstructionLengthSanitisation(TransformPass):
    """
    Checks if quantum instructions are too long and splits if necessary.
    """

    def __init__(self, model: QuantumHardwareModel, target_data: TargetData):
        """
        :param duration_limit: The maximum allowed clock cycles per instruction.

        .. warning::

            The pass will assume that the durations of instructions are sanitised to the
            granularity of the pulse channels. If instructions that do not meet the criteria are
            provided, it might produce incorrect instructions (i.e., instructions that are shorter than
            the clock cycle). This can be enforced using the :class:`InstructionGranularitySanitisation <qat.middleend.passes.transform.InstructionGranularitySanitisation>`
            pass.
        """
        self.duration_limit = target_data.QUBIT_DATA.pulse_duration_max
        self.qubit_sample_time = target_data.QUBIT_DATA.sample_time
        self.resonator_sample_time = target_data.RESONATOR_DATA.sample_time
        self.is_resonator = self._create_resonator_map(model)

    @staticmethod
    def _create_resonator_map(model: QuantumHardwareModel) -> dict[str, bool]:
        """Creates a mapping between physical channel and pulse channel to specify if the
        channel is a resonator or not."""

        is_resonator = {}
        for qubit in model.qubits:
            is_resonator[qubit.physical_channel.id] = False
            is_resonator[qubit.measure_device.physical_channel.id] = True
        return is_resonator

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        """

        new_instructions = []
        for instr in ir.instructions:
            if isinstance(instr, Delay) and instr.duration > self.duration_limit:
                new_instructions.extend(self._batch_delay(instr, self.duration_limit))

            elif (
                isinstance(instr, Pulse)
                and instr.width > self.duration_limit
                and instr.shape == PulseShapeType.SQUARE
            ):
                new_instructions.extend(
                    self._batch_square_pulse(instr, self.duration_limit)
                )

            elif isinstance(instr, CustomPulse) and instr.duration > self.duration_limit:
                new_instructions.extend(
                    self._batch_custom_pulse(instr, self.duration_limit)
                )
            else:
                new_instructions.append(instr)

        ir.instructions = new_instructions
        return ir

    def _batch_delay(self, instruction: Delay, max_duration: float):
        n_instr = int(instruction.duration // max_duration)
        remainder = instruction.duration % max_duration

        batch_instr = []
        for _ in range(n_instr):
            batch_instr.append(Delay(instruction.quantum_targets, time=max_duration))

        if remainder:
            batch_instr.append(Delay(instruction.quantum_targets, time=remainder))

        return batch_instr

    def _batch_square_pulse(self, instruction: Pulse, max_width: float):
        n_instr = int(instruction.width // max_width)
        remainder = instruction.width % max_width

        batch_instr = []
        pulse = deepcopy(instruction)
        pulse.width = max_width
        for _ in range(n_instr):
            batch_instr.append(deepcopy(pulse))

        if remainder:
            pulse = deepcopy(pulse)
            pulse.width = remainder
            batch_instr.append(pulse)

        return batch_instr

    def _batch_custom_pulse(self, instruction: CustomPulse, max_duration: float):
        """Breaks up a custom pulse into multiple custom pulses with none exceeding the
        maximum duration."""

        n_instr = int(instruction.duration // max_duration)
        remainder = instruction.duration % max_duration
        sample_time = (
            self.resonator_sample_time
            if self.is_resonator[instruction.channel.physical_channel.id]
            else self.qubit_sample_time
        )
        max_samples = int(round(max_duration / sample_time))

        batch_instr = []
        for i in range(n_instr):
            pulse = CustomPulse(
                instruction.channel,
                instruction.samples[i * max_samples : (i + 1) * max_samples],
                instruction.ignore_channel_scale,
            )
            batch_instr.append(pulse)

        if remainder:
            num_samples = int(round(remainder / sample_time))
            pulse = CustomPulse(
                instruction.channel,
                instruction.samples[-num_samples:],
                instruction.ignore_channel_scale,
            )
            batch_instr.append(pulse)

        return batch_instr


class SynchronizeTask(TransformPass):
    """Synchronizes all active pulse channels in a task.

    Adds a synchronize to the end of the instruction list for all active pulse channels,
    which is extracted from the :class:`ActivePulseChannelAnalysis` pass. This is useful to
    do before resetting the qubits, as it ensures no qubit is reset while the task is still
    "active", reducing the effects of cross-talk.
    """

    def run(
        self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs
    ) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        active_channels = res_mgr.lookup_by_type(ActiveChannelResults).targets
        if len(active_channels) > 0:
            ir.add(Synchronize(list(active_channels)))
        return ir


class EndOfTaskResetSanitisation(TransformPass):
    """Checks for a reset on each active qubit at the end of a task, and adds Reset
    operations if not found.

    After each shot, it is expected that the qubit is returned to its ground state, ready
    for the next shot. This pass ensures this is the case by checking if the last "active"
    operation on an qubit is a :class:`Reset`, and if not, adds a :class:`Reset` to the end
    of the instruction list. By default, the :class:`Reset` operation will have the drive
    channel of a qubit as its target. However, if the drive channel is not an active
    channel, a different active channel on the qubit will be used its place.

    :class:`Reset` instructions currently sit in a weird place. Their targets are drive
    channels to match the semantics of other quantum instructions (but are instantiated with
    a :class:Qubit). However, like measurements, resets are a qubit-level operation. You
    cannot reset the state of a pulse channel, the state is a property of the qubit! To
    avoid breaking changes, we'll just deal with that for now, but hope to do better in the
    refactored instructions...
    """

    def run(
        self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs
    ) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        """

        active_pulse_channels = res_mgr.lookup_by_type(ActiveChannelResults)
        qubit_map: dict[Qubit, None | bool] = {
            qubit: None for qubit in active_pulse_channels.qubits
        }

        for inst in reversed(ir.instructions):
            if not isinstance(inst, (Pulse, CustomPulse, Acquire, Reset)):
                continue

            elif isinstance(inst, Reset):
                for target in inst.quantum_targets:
                    # if a qubit only sees a reset, its not "active", so ignore
                    qubit = active_pulse_channels.target_map[target]
                    if qubit in qubit_map and qubit_map[qubit] is None:
                        qubit_map[qubit] = True
            else:
                target = active_pulse_channels.target_map[next(iter(inst.quantum_targets))]
                if qubit_map[target] is None:
                    qubit_map[target] = False

            if all([val is not None for val in qubit_map.values()]):
                break

        for qubit, val in qubit_map.items():
            if not val:
                if (
                    target := qubit.get_drive_channel()
                ) not in active_pulse_channels.target_map:
                    # if the drive channel isn't an active channel, choose something else
                    target = active_pulse_channels.from_qubit(qubit)[0]
                ir.add(Reset(target))

        return ir


class ResetsToDelays(TransformPass):
    """
    Transforms :class:`Reset` operations to :class:`Delay`s.

    Note that the delays do not necessarily agree with the granularity of the underlying target machine.
    This can be enforced using the :class:`InstructionGranularitySanitisation`
    pass.
    """

    def __init__(self, target_data: TargetData):
        """
        :param target_data: Target-related information.
        """
        self.passive_reset_time = target_data.QUBIT_DATA.passive_reset_time

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        *args,
        compiler_config: CompilerConfig = None,
        **kwargs,
    ) -> InstructionBuilder:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to store the analysis results.
        :param compiler_config: The compiler configuration.
        """
        active_pulse_channels = res_mgr.lookup_by_type(ActiveChannelResults)

        reset_time = self._get_reset_time(compiler_config, ir)

        new_instructions = []
        for instr in ir.instructions:
            if isinstance(instr, Reset):
                qubit_targets = {
                    active_pulse_channels.target_map[target]
                    for target in instr.quantum_targets
                }

                for qubit in qubit_targets:
                    new_instructions.append(
                        Delay(
                            active_pulse_channels.from_qubit(qubit),
                            time=reset_time,
                        )
                    )

            else:
                new_instructions.append(instr)

        ir.instructions = new_instructions
        return ir

    def _get_reset_time(
        self,
        compiler_config: CompilerConfig,
        ir: InstructionBuilder = None,
    ) -> float:
        """Gets the reset time for a qubit from the compiler configuration, or falls back to
        the passive reset time from the target data."""

        if getattr(compiler_config, "passive_reset_time", None) is not None:
            return compiler_config.passive_reset_time

        if getattr(compiler_config, "repetition_period", None) is not None:
            log.warning(
                "The `repetition_period` in `CompilerConfig` will soon be deprecated. "
                "Please use `passive_reset_time` instead. "
            )
            return self._calculate_reset_time_for_repetition_period(compiler_config, ir)

        return self.passive_reset_time

    def _calculate_reset_time_for_repetition_period(
        self,
        compiler_config: CompilerConfig,
        ir: InstructionBuilder,
    ) -> float:
        """Calculates the reset time for a qubit based on the repetition period in the
        compiler configuration."""

        total_duration = self._get_total_duration(ir)
        reset_time = compiler_config.repetition_period - total_duration

        if reset_time < 0:
            log.warning(
                "The specified `repetition_period` is shorter than the total instruction duration. "
                "Setting the repetition period to equal the total instruction duration."
            )
            return 0.0
        return reset_time

    @staticmethod
    def _get_total_duration(ir: InstructionBuilder) -> float:
        durations: dict[str, float] = defaultdict(float)
        for inst in ir.instructions:
            if isinstance(inst, (Pulse, Acquire, Delay, CustomPulse)):
                # only increment the durations for meaningful instructions
                pulse_chan_id = inst.quantum_targets[0].partial_id()
                durations[pulse_chan_id] += inst.duration
        return max(durations.values())


class EvaluatePulses(TransformPass):
    """Evaluates the amplitudes of :class:`Pulse` instructions, replacing them with a
    :class:`CustomPulse` and accounting for the scale of the pulse channel.

    :class:`Pulse` instructions are defined by (often many) parameters. With the exception
    of specific shapes, they cannot be implemented directly on hardware. Instead, we
    evaluate the waveform at discrete times, and communicate these values to the target.
    This pass evaluates pulses early in the compilation pipeline.

    The :class:`CustomPulse` instructions will have :code:`ignore_channel_scale=True`.
    """

    def __init__(
        self,
        ignored_shapes: list[PulseShapeType] | None = None,
        acquire_ignored_shapes: list[PulseShapeType] | None = None,
        eval_function: callable = evaluate_shape,
    ):
        """
        :param ignored_shapes: A list of pulse shapes that are not evaluated to a custom
            pulse, defaults to `[PulseShapeType.SQUARE]`.
        :param acquire_ignored_shapes: A list of pulse shapes that are not evaluated to a
            custom pulse for :class:`Acquire` filters, defaults to `[]`.
        :param eval_function: Allows a pulse evaluation function to be injected, defaults
            to :meth:`evaluate_shape`.
        """

        ignored_shapes = (
            ignored_shapes if ignored_shapes is not None else [PulseShapeType.SQUARE]
        )
        self.ignored_shapes = (
            ignored_shapes if isinstance(ignored_shapes, list) else [ignored_shapes]
        )
        acquire_ignored_shapes = (
            acquire_ignored_shapes if acquire_ignored_shapes is not None else []
        )
        self.acquire_ignored_shapes = (
            acquire_ignored_shapes
            if isinstance(acquire_ignored_shapes, list)
            else [acquire_ignored_shapes]
        )
        self.evaluate = eval_function

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> InstructionBuilder:
        """:param ir: The list of instructions as an instruction builder."""

        pulses: dict[str, CustomPulse] = dict()

        instructions = []
        for inst in ir.instructions:
            if isinstance(inst, (Pulse, CustomPulse)):
                inst = self.evaluate_waveform(inst, self.ignored_shapes, pulses)
            elif isinstance(inst, Acquire) and inst.filter:
                inst.filter = self.evaluate_waveform(
                    inst.filter, self.acquire_ignored_shapes, pulses
                )
            instructions.append(inst)
        ir.instructions = instructions
        return ir

    def evaluate_waveform(
        self,
        inst: Pulse | CustomPulse,
        ignored_shapes: list[PulseShapeType],
        pulse_lookup: dict[str, CustomPulse],
    ) -> Pulse | CustomPulse:
        """Evaluates the waveform for a :class:`Pulse` or :class:`CustomPulse`, accounting
        for the pulse channel scale."""

        if isinstance(inst, CustomPulse):
            # custom pulses need to be changed to account for scale
            if not inst.ignore_channel_scale:
                target: PulseChannel = inst.channel
                inst.samples = np.asarray(inst.samples) * target.scale
                inst.ignore_channel_scale = True
            return inst

        if isinstance(inst, Pulse):
            if inst.shape not in ignored_shapes:
                # check if the pulse has already been compiled!
                hash_ = self.hash_pulse(inst)
                if hash_ in pulse_lookup:
                    inst = pulse_lookup[hash_]
                    return inst

                # evaluate for non-ignored shapes
                edge = inst.duration / 2.0 - inst.channel.sample_time * 0.5
                samples = int(np.ceil(inst.duration / inst.channel.sample_time - 1e-10))
                t = np.linspace(start=-edge, stop=edge, num=samples)
                pulse_shape = self.evaluate(inst, t, 0.0)
                if not inst.ignore_channel_scale:
                    pulse_shape = pulse_shape * inst.channel.scale
                inst = CustomPulse(inst.channel, pulse_shape, True)
                pulse_lookup[hash] = inst

            elif not inst.ignore_channel_scale:
                # shapes that are ignored are still checked for scale!
                inst.amp *= inst.channel.scale
                inst.ignore_channel_scale = True

            return inst

        raise ValueError(
            f"Expected to see a Pulse or CustomPulse type, got {type(inst)} instead."
        )

    def hash_pulse(self, pulse: Pulse) -> str:
        """Hashs a pulse object."""
        return hash(
            (
                pulse.channel.partial_id,
                pulse.shape,
                pulse.width,
                pulse.amp,
                pulse.phase,
                pulse.drag,
                pulse.rise,
                pulse.amp_setup,
                pulse.scale_factor,
                pulse.zero_at_edges,
                pulse.beta,
                pulse.frequency,
                pulse.internal_phase,
                pulse.std_dev,
                pulse.square_width,
                pulse.ignore_channel_scale,
            )
        )


class ReturnSanitisation(TransformPass):
    """Squashes all :class:`Return` instructions into a single one. Adds a :class:`Return`
    with all acquisitions if none is found."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        returns = [inst for inst in ir.instructions if isinstance(inst, Return)]
        acquires = [inst for inst in ir.instructions if isinstance(inst, Acquire)]

        if returns:
            unique_variables = set(itertools.chain(*[ret.variables for ret in returns]))
            for ret in returns:
                ir._instructions.remove(ret)
        else:
            # If we don't have an explicit return, imply all results.
            unique_variables = set(acq.output_variable for acq in acquires)

        ir.returns(list(unique_variables))
        return ir


class RepeatSanitisation(TransformPass):
    """Adds repeat counts and repetition periods to :class:`Repeat` instructions. If none
    is found, a repeat instruction is added."""

    def __init__(self, model: QuantumHardwareModel, target_data: TargetData):
        """
        :param model: The hardware model contains the default repeat value, defaults to
            None.
        :param target_data: Target-related information.
        """
        self.model = model
        self.target_data = target_data

    def run(
        self,
        ir: InstructionBuilder,
        *args,
        compiler_config: CompilerConfig = None,
        **kwargs,
    ):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""
        num_shots = self.target_data.default_shots
        configured = False
        if compiler_config is not None and compiler_config.repeats is not None:
            configured = True
            num_shots = compiler_config.repeats

        repeats = [inst for inst in ir._instructions if isinstance(inst, Repeat)]
        if repeats:
            for rep in repeats:
                if rep.repeat_count is None:
                    rep.repeat_count = num_shots
                if rep.passive_reset_time is None and rep.repetition_period is None:
                    rep.passive_reset_time = self.target_data.QUBIT_DATA.passive_reset_time
            if not configured:
                num_shots = repeats[0].repeat_count
            if not all([rep.repeat_count == num_shots for rep in repeats]):
                raise ValueError(
                    "Inconsistent repeat_count information found. "
                    + f"Repeat instruction values: {[rep.repeat_count for rep in repeats]}"
                    + (f", CompilerConfig repeats: {num_shots}." if configured else ".")
                )
        else:
            ir.repeat(
                num_shots,
                passive_reset_time=self.target_data.QUBIT_DATA.passive_reset_time,
            )
        return ir


class FreqShiftSanitisation(TransformPass):
    """
    Looks for any active frequency shift pulse channels in the hardware model and adds
    square pulses for the duration.

    .. warning::

        This pass assumes the following, which is achieved via other passes:

        * :class:`Synchronize` instructions have already been lowered to :class:`Delay`
          instructions.
        * Durations are static.
    """

    def __init__(self, model: QuantumHardwareModel):
        """:param model: The hardware model containing the frequency shift channels."""

        self.model = model

    def run(
        self, ir: InstructionBuilder, res_mgr: ResultManager, *args, **kwargs
    ) -> InstructionBuilder:
        """
        :param ir: The QatIR as an instruction builder.
        :param res_mgr: The results manager.
        """

        freq_shift_channels = self.get_freq_shift_channels()
        if len(freq_shift_channels) == 0:
            return ir
        ir = self.add_freq_shift_to_ir(ir, set(freq_shift_channels.keys()))

        if res_mgr.check_for_type(ActiveChannelResults):
            res = res_mgr.lookup_by_type(ActiveChannelResults)
            res.target_map.update(freq_shift_channels)

        return ir

    def get_freq_shift_channels(self) -> dict[PulseChannel, Qubit]:
        """Returns all active frequency shift pulse channels found in the hardware model."""

        channels: dict[PulseChannel, Qubit] = dict()
        for qubit in self.model.qubits:
            try:
                freq_shift_pulse_ch = qubit.get_freq_shift_channel()
                if freq_shift_pulse_ch.active:
                    channels[freq_shift_pulse_ch] = qubit
            except KeyError:
                continue

        return channels

    @staticmethod
    def add_freq_shift_to_ir(
        ir: InstructionBuilder, freq_shift_channels: set[PulseChannel]
    ) -> InstructionBuilder:
        """Adds frequency shift instructions to the instruction builder."""

        durations = defaultdict(list)
        for inst in ir.instructions:
            if isinstance(inst, QuantumInstruction) and (duration := inst.duration) > 0:
                for target in inst.quantum_targets:
                    durations[target].append(duration)

        if len(durations) == 0:
            return ir

        max_duration = np.max([np.sum(duration) for duration in durations.values()])
        if max_duration == 0.0:
            return ir

        # Find the index if the first quantum instruction
        index = next(
            filter(
                lambda tup: True if isinstance(tup[1], QuantumInstruction) else None,
                enumerate(ir._instructions),
            )
        )[0]

        for channel in freq_shift_channels:
            pulse = Pulse(
                channel,
                shape=PulseShapeType.SQUARE,
                amp=channel.amp,
                phase=getattr(channel, "phase", 0),
                width=max_duration,
            )
            ir.insert(pulse, index)
        return ir


class LowerSyncsToDelays(TransformPass):
    """Lowers :class:`Synchronize` instructions to :class:`Delay` instructions with static
    times.

    Increments through the instruction list, keeping track of the cumulative duration.
    When :class:`Synchronize` instructions are encountered, it is replaced with
    :class:`Delay` instructions with timings calculated from the cumulative durations.

    .. warning::

        Any manipulations of the instruction set that will alter the timeline and occur
        after this pass could invalidate the intention of the :class:`Synchronize`
        instruction.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> InstructionBuilder:
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        durations: dict[str, float] = defaultdict(float)
        new_instructions: list[Instruction] = []

        for inst in ir.instructions:
            if isinstance(inst, (Pulse, Acquire, Delay, CustomPulse)):
                # only increment the durations for meaningful instructions
                pulse_chan_id = inst.quantum_targets[0].partial_id()
                durations[pulse_chan_id] += inst.duration
                new_instructions.append(inst)
            elif isinstance(inst, Synchronize):
                # determine the durations for syncs
                targets = inst.quantum_targets
                if (num_targets := len(targets)) <= 1:
                    log.info(
                        f"Synchronize instructions with {num_targets} does not have enough "
                        "targets to synchronise, it will be ignored."
                    )
                    continue
                current_durations = np.asarray(
                    [durations[target.partial_id()] for target in targets]
                )
                max_duration = np.max(current_durations)
                sync_durations = max_duration - current_durations
                delay_instrs = [
                    Delay(target, sync_durations[i])
                    for i, target in enumerate(targets)
                    if sync_durations[i] > 0.0
                ]
                new_instructions.extend(delay_instrs)
                durations.update({target.partial_id(): max_duration for target in targets})
            else:
                # every other instruction is just added to the list
                new_instructions.append(inst)

        ir.instructions = new_instructions
        return ir


class SquashDelaysOptimisation(TransformPass):
    """Looks for consecutive :class:`Delay` instructions on a pulse channel and squashes
    them into a single instruction.

    Because :class:`Synchronize` instructions across multiple pulse channels are used so
    frequently to ensure pulses play at the correct timings, it means we can have sequences
    of many delays. Reducing the number of delays will simplify timing analysis later in
    the compilation.

    :class:`Delay` instructions commute with phase related instructions, so the only
    instructions that separate delays in a meaningful way are: :class:`Pulse`:,
    :class:`CustomPulse` and :class:`Acquire` instructions. We also need to be careful to
    not squash delays that contain a variable time.
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        :param met_mgr: The metrics manager to store the number of instructions after
            optimisation.
        """
        delimiter_types = (
            Acquire,
            Assign,
            CustomPulse,
            Delay,
            Jump,
            Label,
            PhaseReset,
            PhaseSet,
            Pulse,
        )
        accumulated_delays: dict[PulseChannel, float] = defaultdict(float)
        instructions: list[Instruction] = []
        for inst in ir.instructions:
            if isinstance(inst, Delay) and isinstance(inst.time, Number):
                for target in inst.quantum_targets:
                    accumulated_delays[target] += inst.time
            elif isinstance(inst, delimiter_types):
                if isinstance(inst, QuantumInstruction):
                    targets = inst.quantum_targets
                else:
                    targets = accumulated_delays.keys()
                for target in targets:
                    if (time := accumulated_delays[target]) > 0.0:
                        instructions.append(Delay(target, time))
                        accumulated_delays[target] = 0.0
                instructions.append(inst)
            else:
                instructions.append(inst)

        for key, val in accumulated_delays.items():
            if val != 0.0:
                instructions.append(Delay(key, val))
        ir.instructions = instructions

        met_mgr.record_metric(MetricsType.OptimizedInstructionCount, len(instructions))
        return ir


class BatchedShots(TransformPass):
    """Determines how shots should be grouped when the total number exceeds that maximum
    allowed.

    The target machine might have an allowed number of shots that can be executed by a
    single execution call. To execute a number of shots greater than this value, shots can
    be batched, with each batch executed by its own "execute" call on the target machine.
    For example, if the maximum number of shots for a target machine is 2000, but you
    required 4000 shots, then this could be done as [2000, 2000] shots.

    Now consider the more complex scenario where  4001 shots are required. Clearly this can
    be done in three batches. While it is tempting to do this in batches of [2000, 2000, 1],
    for some target machines, specification of the number of shots can only be achieved at
    compilation (as opposed to runtime). Batching as described above would result in us
    needing to compile two separate programs. Instead, it makes more sense to batch the
    shots as three lots of 1334 shots, which gives a total of 4002 shots. The extra two
    shots can just be discarded at run time.

    .. warning::

        This pass makes certain assumptions about the IR, including that there is only
        at most one :class:`Repeat` instruction that contributes to the number of readouts,
        and there is at most one readout per shot. It will change the number of shots in
        the :class:`Repeat` instruction to the decided batch size, and store the total
        required shots in the IR, which is less than ideal. It would be nice to save this
        as an analysis result, but an unfortante side effect of the decoupled nature of
        the middle- and back-end is that the results manager might not necessarily be
        passed along.
    """

    def __init__(self, target_data: TargetData):
        """
        :param target_data: Target-related information.
        """
        self.max_shots = target_data.max_shots

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        repeats = [inst for inst in ir.instructions if isinstance(inst, Repeat)]
        if len(repeats) == 0:
            return ir

        if len(repeats) > 1:
            log.warning(
                "Multiple repeat instructions found. Using only the first for batching of "
                "shots."
            )

        repeat = repeats[0]
        num_shots = repeat.repeat_count
        num_batches = int(np.ceil(num_shots / self.max_shots))
        shots_per_batch = int(np.ceil(num_shots / num_batches))

        if num_batches > 1:
            log.info(
                f"The number of shots {num_shots} exceeds the maximum allowed on the "
                f"target. Batching as {num_batches} batches of {shots_per_batch} shots."
            )

        repeat.repeat_count = shots_per_batch
        ir.shots = num_shots
        ir.compiled_shots = shots_per_batch
        return ir


class ScopeSanitisation(TransformPass):
    """Bubbles up all sweeps and repeats to the beginning of the list and adds delimiter
    instructions to the repeats and sweeps signifying the end of their scopes.

    .. warning::

        This pass is intended for use with legacy builders that do not use
        :class:`EndRepeat` and :class:`EndSweep` instructions. It will add a delimiter
        instruction for each :class:`Repeat` and :class:`Sweep` found in the IR.
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """:param ir: The list of instructions stored in an :class:`InstructionBuilder`."""

        tail, head = partition(
            lambda inst: isinstance(inst, (Sweep, Repeat)), ir.instructions
        )
        tail, head = list(tail), list(head)

        delimiters = [
            EndSweep() if isinstance(inst, Sweep) else EndRepeat() for inst in head
        ]

        ir.instructions = head + tail + delimiters[::-1]
        return ir


class DeviceUpdateSanitisation(TransformPass):
    """
    Duplicate DeviceUpdate instructions upsets the device injection mechanism, which causes corruption
    of the HW model.

    In fact, a DeviceInjector is currently 1-1 associated with a DeviceUpdate instruction. When multiple
    DeviceUpdate instructions (sequentially) inject the same "target", the first DeviceInjector assigns the
    (correct) value of the attribute (on the target) to the revert_value. At this point the HW model (or any
    other target kind) is dirty, and any subsequent DeviceInjector updater would surely assign
    the (wrong, usually a placeholder `Variable`) to its revert_value. This results in a corrupt HW model whereby
    reversion wouldn't have the desired effect.

    This pass is a (lazy) fix, which is to analyse when such cases happen and eliminate duplicate DeviceUpdate
    instructions that target THE SAME "attribute" on THE SAME "target" with THE SAME variable.
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        target_attr2value = defaultdict(list)
        for inst in ir.instructions:
            if isinstance(inst, DeviceUpdate):
                # target already validated to be a QuantumComponent during initialisation
                if not hasattr(inst.target, inst.attribute):
                    raise ValueError(
                        f"Attempting to assign {inst.value} to non existing attribute {inst.attribute}"
                    )

                if isinstance(inst.value, Variable):
                    target_attr2value[(inst.target, inst.attribute)].append(inst.value)

        for (target, attr), values in target_attr2value.items():
            if len(values) > 1:
                log.warning(
                    f"Multiple DeviceUpdate instructions attempting to update the same attribute '{attr}' on {target}"
                )

            unique_values = set(values)
            if len(unique_values) > 1:
                raise ValueError(
                    f"Cannot update the same attribute '{attr}' on {target} with distinct values {unique_values}"
                )

        new_instructions = []
        for inst in ir.instructions:
            if isinstance(inst, DeviceUpdate) and isinstance(inst.value, Variable):
                if next(iter(target_attr2value[(inst.target, inst.attribute)]), None):
                    target_attr2value[(inst.target, inst.attribute)].clear()
                    new_instructions.append(inst)
            else:
                new_instructions.append(inst)

        ir.instructions = new_instructions

        return ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.pass_base import PassManager
from qat.core.result_base import ResultManager
from qat.executables import Executable
from qat.model.target_data import AbstractTargetData
from qat.pipelines.base import AbstractPipeline
from qat.pipelines.pipeline import CompilePipeline
from qat.pipelines.updateable import UpdateablePipeline
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.hardware_models import QuantumHardwareModel

from .flattener import SweepFlattener
from .passes import FrequencyAssignSanitisation


class CompileSweepPipeline(AbstractPipeline):
    """A pipeline that shims support for compiling instruction builders with sweeps and
    devices assigns by making repeated calls to an underlying pipeline.

    The underlying pipeline must be updateable. This is done because applying device assigns
    can change properties of the hardware model, and the pipeline must be rebuilt to ensure
    it is not invalidated. Furthermore, even if device assigns aren't used, sweeps are
    often done side-by-side with changes to the hardware model, so it is useful to ensure
    its rebuilt as a precaution.

    Eventually the risks associated with device assigns can be mitigated by mapping device
    assigns onto instructions in the IR, where possible. For example, a device assign that
    changes the frequency of a channel can be mapped onto a frequency set instruction.
    Similar behaviour could implemented for pulse channel scales too, with an e.g.
    "set scale" instruction, or appropriate mapping onto pulse amplitudes.
    """

    def __init__(
        self,
        base_pipeline: UpdateablePipeline,
        preprocessing_pipeline: PassManager | None = None,
    ):
        """
        :param base_pipeline: The underlying pipeline that is used to compile each
            instance of the sweep.
        :param preprocessing_pipeline: An optional preprocessing pipeline that runs on the
            IR before sweeps are flattened out.
        """
        if not base_pipeline.is_subtype_of(UpdateablePipeline):
            raise TypeError(
                "The base pipeline must be an UpdateablePipeline to support dynamic "
                "hardware models."
            )
        if not base_pipeline.is_subtype_of(CompilePipeline):
            raise TypeError("CompileSweepPipeline can only wrap CompilePipelines.")

        self._base_pipeline = base_pipeline
        self._preprocessing_pipeline = (
            preprocessing_pipeline
            if preprocessing_pipeline is not None
            else self._build_preprocessing_pipeline()
        )

    def compile(
        self,
        program: str | InstructionBuilder,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ):
        """Compile a builder with sweeps and device assigns against the base pipeline.

        :param program: The program to compile, which might include an instruction builder
            with sweeps.
        :param compiler_config: Optional compiler configuration to use for this compile
            call. If not provided, the compiler configuration from the pipeline will be
            used.
        :return: A batched executable containing all instances of the sweeps, along with
            the metrics from the final compile call.
        """
        if not isinstance(program, InstructionBuilder):
            return self._base_pipeline.compile(program, compiler_config)

        # done as precaution, since external forces might change the model between compile
        # calls
        self._rebuild_pipeline()
        self._run_preprocessing_pipeline(program)

        flattener = SweepFlattener(program)
        sweep_shape = flattener.sweep_sizes
        sweep_instances = flattener.create_flattened_builders()

        executables = []
        for sweep_instance in sweep_instances:
            if sweep_instance.has_device_assigns:
                executable, metrics = self._compile_with_device_assigns(
                    sweep_instance.builder,
                    sweep_instance.device_assigns,
                    compiler_config,
                    **kwargs,
                )
            else:
                executable, metrics = self._base_pipeline.compile(
                    sweep_instance.builder, compiler_config, **kwargs
                )
            executables.append(executable)

        executable = self._combine_executables(executables, sweep_shape)
        return executable, metrics

    def is_subtype_of(self, cls):
        return isinstance(self, cls) or self._base_pipeline.is_subtype_of(cls)

    @property
    def name(self) -> str:
        return self._base_pipeline.name

    @property
    def model(self) -> QuantumHardwareModel:
        return self._base_pipeline.model

    @property
    def target_data(self) -> AbstractTargetData:
        return self._base_pipeline.target_data

    @staticmethod
    def _build_preprocessing_pipeline() -> PassManager:
        """Builds a preprocessing pipeline that runs on the IR before sweeps are flattened
        out, such as resolving device assigns onto other instructions. Currently only
        sanitises frequencies, but more passes can be added as needed."""

        return PassManager() | FrequencyAssignSanitisation()

    def _run_preprocessing_pipeline(self, ir: InstructionBuilder) -> InstructionBuilder:
        """Runs the preprocessing pipeline on the IR, if one is set."""
        return self._preprocessing_pipeline.run(ir, ResultManager(), MetricsManager())

    def _rebuild_pipeline(self):
        """Rebuilds the underlying pipelines with the same model to ensure any changes to
        model do not invalidate the pipeline."""
        self._base_pipeline.update(model=self.model)

    def _compile_with_device_assigns(
        self,
        builder: InstructionBuilder,
        device_assigns,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ) -> tuple[Executable, MetricsManager]:
        """Makes required changes to the hardware model using the device assigns,
        compiles the builder, then restores the hardware model to its original state."""
        with device_assigns.apply():
            self._rebuild_pipeline()
            return self._base_pipeline.compile(builder, compiler_config, **kwargs)

    def _combine_executables(self, executables: list[Executable], sweep_shape: tuple[int]):
        """Combines the programs within the executables into a single executable."""

        programs = []
        for executable in executables:
            programs.extend(executable.programs)

        acquire_data = executables[0].acquires
        for acquire in acquire_data.values():
            # This is done intentionally: legacy qat returns with an extra dimension (list)
            # if sweeps aren't used... so its like [[result1, result2, ...]]
            if len(sweep_shape) == 0:
                sweep_shape = (1,)
            acquire.shape = tuple(list(sweep_shape) + list(acquire.shape))

        return Executable(
            programs=programs,
            acquires=acquire_data,
            assigns=executables[0].assigns,
            returns=executables[0].returns,
            calibration_id=executables[0].calibration_id,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from collections import defaultdict
from dataclasses import asdict

from qblox_instruments import Cluster
from qblox_instruments.qcodes_drivers.module import Module
from qblox_instruments.qcodes_drivers.sequencer import Sequencer

from qat.backend.qblox.acquisition import Acquisition
from qat.backend.qblox.codegen import QbloxPackage
from qat.backend.qblox.config.constants import Constants
from qat.backend.qblox.config.helpers import (
    QcmConfigHelper,
    QcmRfConfigHelper,
    QrmConfigHelper,
    QrmRfConfigHelper,
)
from qat.backend.qblox.execution import QbloxProgram
from qat.instrument.base import CompositeInstrument, LeafInstrument
from qat.purr.compiler.devices import ChannelType
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxLeafInstrument(LeafInstrument):
    def __init__(
        self,
        id: str,
        name: str,
        address: str = None,
        ref_source: str = None,
    ):
        super().__init__(id=id, name=name, address=address)
        self.ref_source = ref_source or "internal"

        self._driver: Cluster = None
        self._modules: dict[Module, bool] = {}
        self._id2seq: dict[str, Sequencer] = {}

    def _reset_modules(self):
        # TODO - Qblox bug: Hard reset clutters sequencer connections with conflicting defaults
        # TODO - This is a temporary workaround until Qblox fixes the issue

        modules = [mod for mod, is_dirty in self._modules.items() if is_dirty]

        for mod in modules:
            log.debug(f"Resetting sequencer connections for module {mod}")
            mod.disconnect_outputs()
            if mod.is_qrm_type:
                mod.disconnect_inputs()
            self._modules[mod] = False  # Mark as clean

    @property
    def driver(self):
        return self._driver

    @property
    def id2seq(self):
        return self._id2seq

    @property
    def modules(self):
        return self._modules

    def configure(self, package: QbloxPackage):
        module: Module = getattr(self._driver, f"module{package.slot_idx}")
        sequencer = module.sequencers[package.seq_idx]

        seq_config, mod_config = package.seq_config, package.mod_config
        if module.is_qcm_type:
            if module.is_rf_type:
                config_helper = QcmRfConfigHelper(mod_config, seq_config)
            else:
                config_helper = QcmConfigHelper(mod_config, seq_config)
        elif module.is_qrm_type:
            if module.is_rf_type:
                config_helper = QrmRfConfigHelper(mod_config, seq_config)
            else:
                config_helper = QrmConfigHelper(mod_config, seq_config)
        else:
            raise ValueError(f"Unknown module type {module.module_type}")

        try:
            log.debug(f"Configuring module {module}, and sequencer {sequencer}")
            config_helper.configure(module, sequencer)
            sequence = asdict(package.sequence)
            sequencer.sequence(sequence)
        finally:
            self._modules[module] = True  # Mark as dirty
            self._id2seq[package.pulse_channel_id] = sequencer

    def connect(self):
        if self._driver is None or not Cluster.is_valid(self._driver):
            self._driver: Cluster = Cluster(name=self.name, identifier=self.address)
            self._driver.reset()
            self._driver.reference_source(self.ref_source)
            self._modules = {m: True for m in self._driver.get_connected_modules().values()}
            self.is_connected = True

        log.info(self._driver.get_system_status())

    def disconnect(self):
        if self._driver is not None:
            try:
                self._driver.close()
                self._driver = None
                self._modules.clear()
                self.is_connected = False
            except BaseException as e:
                log.warning(
                    f"Failed to close instrument ID: {self.id} at: {self.address}\n{str(e)}"
                )

    def setup(self, program: QbloxProgram):
        try:
            self._id2seq.clear()
            self._reset_modules()
            for pkg in program.packages.values():
                self.configure(pkg)
        except BaseException as e:
            self._id2seq.clear()
            self._reset_modules()
            raise e

    def playback(self):
        if not any(self._id2seq):
            raise ValueError("No allocations found. Install packages and configure first")

        results: dict[str, list[Acquisition]] = defaultdict(list)
        try:
            for pulse_channel_id, sequencer in self._id2seq.items():
                sequencer.sync_en(True)

            for pulse_channel_id, sequencer in self._id2seq.items():
                sequencer.arm_sequencer()

            for pulse_channel_id, sequencer in self._id2seq.items():
                sequencer.start_sequencer()

            for pulse_channel_id, sequencer in self._id2seq.items():
                if ChannelType.macq.name in pulse_channel_id:
                    status_obj = sequencer.get_sequencer_status()
                    log.debug(f"Sequencer status - {sequencer}: {status_obj}")
                    if acquisitions := sequencer.get_acquisitions():
                        # TODO - 60 min tops, make it dynamic by involving flow-aware timeline duration
                        # Only wait if you sequencer is expected to have acquisitions
                        # TODO - Precise expectation of acquisitions should come from higher up
                        sequencer.get_acquisition_status(timeout=60)

                    for name in acquisitions:
                        sequencer.store_scope_acquisition(name)

                    # (re)fetch the lot after having stored scope acquisition
                    acquisitions = sequencer.get_acquisitions()

                    integ_length = sequencer.integration_length_acq()
                    start, end = (
                        0,
                        min(
                            integ_length,
                            Constants.MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS,
                        ),
                    )
                    for name, acquisition in acquisitions.items():
                        acquisition = Acquisition.model_validate(acquisition)
                        acquisition.name = name

                        scope_data = acquisition.acquisition.scope
                        scope_data.path0.data = scope_data.path0.data[start:end]
                        scope_data.path1.data = scope_data.path1.data[start:end]

                        integ_data = acquisition.acquisition.bins.integration
                        integ_data.path0 /= integ_length
                        integ_data.path1 /= integ_length

                        results[pulse_channel_id].append(acquisition)

            return results
        finally:
            for pulse_channel_id, sequencer in self._id2seq.items():
                sequencer.sync_en(False)
                if ChannelType.macq.name in pulse_channel_id:
                    sequencer.delete_acquisition_data(all=True)

            self._id2seq.clear()
            self._reset_modules()


class QbloxCompositeInstrument(CompositeInstrument[QbloxLeafInstrument]):
    """
    Composing Qblox instruments can be achieved by 2 methods:
    1- Daisy-chaining the REF_out of one cluster's CMM to REF_in of the next cluster's CMM.
       This is like a linked list pattern where the first cluster is **allowed** to have both "external"
       or "internal" reference source config, but subsequent clusters **must** have their reference
       source set as "external".
    2- Distribute the clock from a common source to all the REF_in of all clusters' CMM modules.
       This is like a star pattern where an external reference clock is distributed to all the clusters
       in the fleet

    Historically, both methods have been supported, but recent FW versions ditched the first method and only
    support the second method. Regardless of the method followed, this abstraction remains oblivious.
    """

    pass


LiveQbloxInstrument = QbloxLeafInstrument | QbloxCompositeInstrument



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
# ruff: noqa F401
from warnings import warn


from qat.backend.waveform.executable import (
    PositionalAcquireData as PositionalAcquireData,
)
from qat.backend.waveform.executable import (
    WaveformChannelData as WaveformV1ChannelData,
)
from qat.backend.waveform.executable import (
    WaveformProgram as WaveformV1Program,
)

warn(
    "WaveformV1 support is deprecated and will be removed in v4.",
    DeprecationWarning,
    stacklevel=2,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from qat.model.target_data import TargetData


class Constants:
    MIN_GAIN = -pow(2, 15)
    """Minimum gain in Q1ASM programs."""
    MAX_GAIN = pow(2, 15) - 1
    """Maximum gain in Q1ASM programs."""
    MAX_GAIN_SIZE = pow(2, 16) - 1
    """Max size of gain in Q1ASM programs."""
    MIN_OFFSET = -pow(2, 15)
    """Minimum offset in Q1ASM programs."""
    MAX_OFFSET = pow(2, 15) - 1
    """Maximum offset in Q1ASM programs."""
    MAX_OFFSET_SIZE = pow(2, 16) - 1
    """Max size of offset in Q1ASM programs."""
    MAX_WAIT_TIME = pow(2, 16) - 4
    """Max size of wait instruction immediates in Q1ASM programs. Max value allowed by
    assembler is 2**16-1, but this is the largest that is a multiple of 4 ns."""
    REGISTER_SIZE = pow(2, 32) - 1
    """Size of registers in Q1ASM programs."""
    NCO_MIN_FREQ = -500e6
    """Minimum NCO frequency"""
    NCO_MAX_FREQ = 500e6
    """Maximum NCO frequency"""
    NCO_MAX_PHASE_STEPS = int(1e9)
    """Maximum NCO phase steps"""
    NCO_PHASE_STEPS_PER_DEG = 1e9 / 360
    """The number of steps per degree for NCO phase instructions arguments."""
    NCO_FREQ_STEPS_PER_HZ = 4.0
    """The number of steps per Hz for the NCO set_freq instruction."""
    NCO_FREQ_LIMIT_STEPS = 2e9
    """The maximum and minimum frequency expressed in steps for the NCO set_freq instruction.
    For the minimum we multiply by -1."""
    NUMBER_OF_SEQUENCERS_QCM = 6
    """Number of sequencers supported by a QCM/QCM-RF in the latest firmware."""
    NUMBER_OF_SEQUENCERS_QRM = 6
    """Number of sequencers supported by a QRM/QRM-RF in the latest firmware."""
    NUMBER_OF_REGISTERS: int = 64
    """Number of registers available in the Qblox sequencers."""
    MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS: int = 16384
    """Maximal amount of scope trace acquisition datapoints returned."""
    MAX_TOTAL_BINNED_ACQUISITIONS: int = 589824
    MAX_012_BINNED_ACQUISITIONS: int = 1 << 17
    MAX_345_BINNED_ACQUISITIONS: int = 1 << 16
    """Each sequencer has 131072 bins with some caveats. If you are using all sequencers
    you are limited to 131k bins for the first three sequencers and 65536 bins for the final
    three to a maximum number of 589824 bins."""
    MIN_ACQ_INTEGRATION_LENGTH = 4
    MAX_ACQ_INTEGRATION_LENGTH = (1 << 24) - 4
    """Minimum and maximum integration lengths"""
    MIN_ACQ_THRESHOLD = -((1 << 24) - 4)
    MAX_ACQ_THRESHOLD = (1 << 24) - 4
    """Minimum and maximum thresholds for the thresholded acquisition"""
    MAX_SAMPLE_SIZE_WAVEFORMS: int = 16384
    """Maximal amount of samples in the waveforms to be uploaded to a sequencer."""
    GRID_TIME = 4  # ns
    """
    Clock period of the sequencers. All time intervals used must be multiples of this value.
    """
    LOOP_UNROLL_THRESHOLD = 4
    """
    Size above which loops have tolerable overhead
    """
    MIN_QCM_OFFSET_V = -2.5
    """
    Minimum offset for QCM
    """
    MAX_QCM_OFFSET_V = 2.5
    """
    Maximum offset for QCM
    """
    MIN_QCM_RF_OFFSET_MV = -84
    """
    Minimum offset for QCM-RF
    """
    MAX_QCM_RF_OFFSET_MV = 73
    """
    Maximum offset for QCM-RF
    """
    MIN_QRM_OFFSET_V = -0.09
    """
    Minimum offset for QRM
    """
    MAX_QRM_OFFSET_V = 0.09
    """
    Maximum offset for QRM
    """
    MIN_QRM_RF_OFFSET_V = -0.09
    """
    Minimum offset for QRM-RF
    """
    MAX_QRM_RF_OFFSET_V = 0.09
    """
    Maximum offset for QRM-RF
    """


# TODO: Populate QbloxTargetData: COMPILER-731
class QbloxTargetData(TargetData):
    pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from qat.engines.waveform_v1 import EchoEngine
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import ExecutePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime import SimpleRuntime
from qat.runtime.results_pipeline import get_default_results_pipeline

log = get_default_logger()


class EchoExecutePipeline(UpdateablePipeline):
    """A pipeline that executes :class:`Executable <qat.executables.Executable>` with
    :class:`WaveformV1Program <qat.backend.waveform_v1.executable.WaveformV1Program>`
    packages using the :class:`EchoEngine`, which simply passes through the waveform
    buffers.

    .. warning::

        This pipeline is for execution purposes only and does not compile programs. Please
        select an appropriate compilation pipeline if you wish to compile programs before
        execution.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None,
        engine: None = None,
    ) -> ExecutePipeline:
        if engine is not None:
            log.warning(
                "An engine was provided to the EchoPipeline, but it will be ignored. "
                "The EchoEngine is used directly."
            )

        target_data = target_data if target_data is not None else TargetData.default()
        results_pipeline = get_default_results_pipeline(model)
        return ExecutePipeline(
            model=model,
            target_data=target_data,
            runtime=SimpleRuntime(engine=EchoEngine(), results_pipeline=results_pipeline),
            name=config.name,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.backend.waveform.codegen import PydWaveformBackend as PydWaveformBackend
from qat.backend.waveform.executable import (
    PositionalAcquireData as PositionalAcquireData,
    WaveformChannelData as WaveformChannelData,
    WaveformProgram as WaveformProgram,
)

WaveformBackend = PydWaveformBackend



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
import random
from enum import Enum, auto
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from qat.purr.backends.utilities import get_axis_map
from qat.purr.compiler.devices import (
    ChannelType,
    PhysicalBaseband,
    PhysicalChannel,
    Qubit,
    QubitCoupling,
    Resonator,
)
from qat.purr.compiler.emitter import QatFile
from qat.purr.compiler.execution import QuantumExecutionEngine, SweepIterator
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import AcquireMode, PostProcessing
from qat.purr.compiler.interrupt import Interrupt, NullInterrupt


def apply_setup_to_hardware(
    hw,
    qubit_count: int = 4,
    connectivity: Optional[List[Tuple[int, int]]] = None,
    add_direction_couplings: bool = False,
):
    """Apply the default echo hardware setup to the passed-in hardware."""
    qubit_devices = []
    resonator_devices = []
    channel_index = 1
    for primary_index in range(qubit_count):
        bb1 = PhysicalBaseband(f"LO{channel_index}", 5.5e9)
        bb2 = PhysicalBaseband(f"LO{channel_index + 1}", 8.5e9)
        hw.add_physical_baseband(bb1, bb2)

        ch1 = PhysicalChannel(f"CH{channel_index}", 1.0e-9, bb1, 1)
        ch2 = PhysicalChannel(
            f"CH{channel_index + 1}", 1.0e-9, bb2, 1, acquire_allowed=True
        )
        hw.add_physical_channel(ch1, ch2)

        resonator = Resonator(f"R{primary_index}", ch2)
        resonator.create_pulse_channel(ChannelType.measure, frequency=8.5e9)
        resonator.create_pulse_channel(ChannelType.acquire, frequency=8.5e9)

        # As our main system is a ring architecture we just attach every qubit in the
        # ring to the one on either side.
        # 2 has a connection to 1 and 3. This number wraps around, so we also have
        # 10-0-1 linkages.
        qubit = Qubit(primary_index, resonator, ch1)
        qubit.create_pulse_channel(ChannelType.drive, frequency=5.5e9)

        qubit_devices.append(qubit)
        resonator_devices.append(resonator)
        channel_index = channel_index + 2

    # TODO: For backwards compatability cross resonance pulse channels are fully
    #   connected but coupled qubits are only in a ring architecture. I think it would be
    #   more approriate for cross resonace channels to also be a ring architecture but
    #   that can be done in a later PR.
    qubits_by_index = {qb.index: qb for qb in qubit_devices}
    if connectivity is None:
        connectivity = generate_connectivity(
            Connectivity.Ring, qubit_count=len(qubit_devices)
        )

    for connection in connectivity:
        left_index, right_index = connection
        qubit_left = qubits_by_index.get(left_index)
        qubit_right = qubits_by_index.get(right_index)
        qubit_right.create_pulse_channel(
            auxiliary_devices=[qubit_left],
            channel_type=ChannelType.cross_resonance,
            frequency=5.5e9,
            scale=50,
        )
        qubit_right.create_pulse_channel(
            auxiliary_devices=[qubit_left],
            channel_type=ChannelType.cross_resonance_cancellation,
            frequency=5.5e9,
            scale=0.0,
        )
        qubit_left.create_pulse_channel(
            auxiliary_devices=[qubit_right],
            channel_type=ChannelType.cross_resonance,
            frequency=5.5e9,
            scale=50,
        )
        qubit_left.create_pulse_channel(
            auxiliary_devices=[qubit_right],
            channel_type=ChannelType.cross_resonance_cancellation,
            frequency=5.5e9,
            scale=0.0,
        )
        qubit_left.add_coupled_qubit(qubit_right)
        qubit_right.add_coupled_qubit(qubit_left)

    if add_direction_couplings:
        for connection in connectivity:
            hw.qubit_direction_couplings.append(
                QubitCoupling(connection, quality=random.randrange(0, 100))
            )
            hw.qubit_direction_couplings.append(
                QubitCoupling(
                    (connection[1], connection[0]), quality=random.randrange(0, 100)
                )
            )

    hw.add_quantum_device(*qubit_devices, *resonator_devices)
    hw.is_calibrated = True

    return hw


def generate_connectivity(con_type, qubit_count):
    if con_type == Connectivity.Ring:
        if qubit_count == 1:
            return []
        if qubit_count == 2:
            return [(0, 1)]
        return [(i, (i + 1) % qubit_count) for i in range(qubit_count)]
    return None


def add_direction_couplings_to_hardware(model, connectivity):
    for edge in connectivity:
        model.qubit_direction_couplings.append(
            QubitCoupling(edge, quality=random.randrange(1, 20))
        )
    return model


class Connectivity(Enum):
    Ring = auto()


def get_default_echo_hardware(
    qubit_count=4,
    connectivity: Optional[Union[Connectivity, List[Tuple[int, int]]]] = None,
    add_direction_couplings: bool = False,
) -> "QuantumHardwareModel":
    """
    Generate a default echo target optionally providing the type of connectivity. Either you pass a
    pre-defined connectivity as defined in the Connectivity enum or a specific connectivity list of
    which qubits connect to which.
    """
    model = QuantumHardwareModel()
    if isinstance(connectivity, Connectivity):
        connectivity = generate_connectivity(connectivity, qubit_count)

    return apply_setup_to_hardware(
        model, qubit_count, connectivity, add_direction_couplings=add_direction_couplings
    )


class EchoEngine(QuantumExecutionEngine):
    """
    A target machine that just returns default values. Primarily used for testing
    and no-backend situations.
    """

    def run_calibrations(self, qubits_to_calibrate=None):
        pass

    def optimize(self, instructions):
        instructions = super().optimize(instructions)
        for instruction in instructions:
            if isinstance(instruction, PostProcessing):
                acq = instruction.acquire
                if acq.mode == AcquireMode.INTEGRATOR:
                    acq.mode = self.model.default_acquire_mode

        return instructions

    def _execute_on_hardware(
        self,
        sweep_iterator: SweepIterator,
        package: QatFile,
        interrupt: Interrupt = NullInterrupt(),
    ) -> Dict[str, np.ndarray]:
        results = {}
        while not sweep_iterator.is_finished():
            sweep_iterator.do_sweep(package.instructions)

            metadata = {"sweep_iteration": sweep_iterator.get_current_sweep_iteration()}
            interrupt.if_triggered(metadata, throw=True)

            position_map = self.create_duration_timeline(package.instructions)
            pulse_channel_buffers = self.build_pulse_channel_buffers(position_map, True)
            buffers = self.build_physical_channel_buffers(pulse_channel_buffers)
            aq_map = self.build_acquire_list(position_map)

            repeats = package.repeat.repeat_count
            for channel_id, aqs in aq_map.items():
                for aq in aqs:
                    # just echo the output pulse back for now
                    response = buffers[aq.physical_channel.full_id()][
                        aq.start : aq.start + aq.samples
                    ]
                    if aq.mode != AcquireMode.SCOPE:
                        if repeats > 0:
                            response = np.tile(response, repeats).reshape((repeats, -1))

                    response_axis = get_axis_map(aq.mode, response)
                    for pp in package.get_pp_for_variable(aq.output_variable):
                        response, response_axis = self.run_post_processing(
                            pp, response, response_axis
                        )

                    var_result = results.setdefault(
                        aq.output_variable,
                        np.empty(
                            sweep_iterator.get_results_shape(response.shape),
                            response.dtype,
                        ),
                    )
                    sweep_iterator.insert_result_at_sweep_position(var_result, response)

        return results



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from typing import Optional

from compiler_config.config import MetricsType
from pydantic import BaseModel, Field, ValidationError, model_validator

snake_cased_flags = [
    val.snake_case_name()
    for val in list(dict(MetricsType.__members__).values())
    if not val.is_composite()
]


class MetricsManager(BaseModel):
    """Pydantic version based on qat.purr.compiler.metrics.py elements."""

    enabled_metrics: Optional[MetricsType] = Field(
        default=MetricsType.Experimental, repr=False, exclude=True
    )

    optimized_circuit: Optional[str] = Field(default=None)
    optimized_instruction_count: Optional[int] = Field(default=None)
    physical_qubit_indices: Optional[list[int]] = Field(default=None)

    @model_validator(mode="before")
    def validate_all_fields_exist(cls, value):
        """Validate that all expected MetricsType flags have defined fields."""
        missing_fields = [val for val in snake_cased_flags if val not in cls.model_fields]
        if any(missing_fields):
            raise ValidationError(
                f"MetricsManager is missing type hints for {', '.join(missing_fields)}."
            )
        return value

    def __init__(
        self, enabled_metrics: Optional[MetricsType] = MetricsType.Experimental, **kwargs
    ):
        super().__init__(enabled_metrics=enabled_metrics, **kwargs)

    def enable(self, enabled_metrics: MetricsType, overwrite=False):
        """
        Enable these sets of metrics for collection. If overwrite is True then the
        passed-in values will overwrite existing ones.
        """
        if enabled_metrics is None:
            return

        if overwrite:
            self.enabled_metrics = enabled_metrics
        else:
            self.enabled_metrics = self.enabled_metrics | enabled_metrics

    def enable_metrics(self, enabled_metrics=None, overwrite=True):
        """
        Enables the set of metrics in the current collection. If overwrite is set to
        true, or there are no compilation metrics it'll create a new collection, if
        overwrite is false it'll enable these metrics in the currently-active
        collection.
        """
        if enabled_metrics is None:
            # log.warning("Attempted to enable metrics with no value. Defaulting.")
            enabled_metrics = MetricsType.Experimental

        # log.info(f"Enabling metrics with flags: {str(enabled_metrics)}")
        self.enable(enabled_metrics, overwrite)

    def are_enabled(self, metric: MetricsType):
        return self.enabled_metrics is not None and metric in self.enabled_metrics

    def record_metric(self, metric: MetricsType, value):
        if not self.are_enabled(metric):
            return

        setattr(self, metric.snake_case_name(), value)

    def get_metric(self, metric: MetricsType):
        return getattr(self, metric.snake_case_name())

    def as_dict(self):
        """Generates a dictionary of the valid metrics."""
        return self.model_dump()

    def merge(self, other: "MetricsManager"):
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Argument other must be of type MetricsManager, not of type: {type(other)}"
            )
        for metric in snake_cased_flags:
            if self.get_metric(metric) is None:
                self.record_metric(metric, other.get_metric(metric))



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from abc import ABC
from typing import Literal

import numpy as np

from qat.ir.gates.base import GateBase
from qat.ir.gates.gates_1q import Gate1Q, Rz
from qat.ir.gates.gates_2q import Gate2Q


class NativeGate(GateBase, ABC):
    """Marks a gate as a native gate."""


class Z_phase(Rz, NativeGate):
    r"""A native implementation of :math:`R_{z}(\theta)` gate via a phase shift.

    Derived from the :class:`Rz` gate as its mathematically no different. However, this
    class is used to mark it as a :class:`NativeGate` for hardware that support virtual-Z
    gates.

    :param qubit: Target qubit index.
    :param theta: Rotation angle
    """

    inst: Literal["Z_phase"] = "Z_phase"


class X_pi_2(Gate1Q, NativeGate):
    r"""A native :math:`X(\pi/2)` gate.

    Mathematically equivalent to :math:`R_{x}(\pi/2)`, but semantically represents part of
    a decomposed gate.

    :param qubit: Index for the qubit that the gate acts on.
    """

    inst: Literal["X_pi_2"] = "X_pi_2"

    @property
    def matrix(self):
        return (1 / np.sqrt(2)) * np.array([[1, -1j], [-1j, 1]])


class ZX_pi_4(Gate2Q, NativeGate):
    r"""A native :math:`ZX(\pi/4)` gate.

    .. warning:: While this gate mathematically represents a :math:`ZX(\pi/4)` gate, it
        is frequently used as a block within an :class:`ECR` gate, and is calibrated for
        this use. Using this gate on its own might yield unexpected results.

    :param qubit1: Target qubit for the Z gate.
    :param qubit2: Target qubit for the X gate.
    """

    inst: Literal["ZX_pi_4"] = "ZX_pi_4"

    @property
    def matrix(self):
        cos = np.cos(np.pi / 8)
        sin = np.sin(np.pi / 8)
        return np.array(
            [
                [cos, -1j * sin, 0, 0],
                [-1j * sin, cos, 0, 0],
                [0, 0, cos, 1j * sin],
                [0, 0, 1j * sin, cos],
            ]
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import numpy as np

from qat.engines.native import NativeEngine
from qat.executables import AbstractProgram
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class ZeroEngine(NativeEngine):
    """An engine primarily used for testing that simply returns zero readout.

    The engine is designed to be applicable to all programs. It just populates the readout
    data for each output variable with zeros.
    """

    def execute(self, program: AbstractProgram, **kwargs) -> dict[str, np.ndarray]:
        """Execute an :class:`AbstractProgram`, returning zeros for all readouts.

        :param package: The compiled executable containing acquisitions.
        :returns: The zero readout results.
        """
        results = {}
        for acquire, shape in program.acquire_shapes.items():
            results[acquire] = np.zeros(shape, dtype=np.complex128)
        return results



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.core.pass_base import ValidationPass
from qat.ir.qat_ir import QatIR


class NoMidCircuitMeasurements(ValidationPass):
    r"""Checks that a circuit has no mid-circuit measurements.

    Looks for :class:`Measure` instructions on a qubit that is followed by any gate- or
    pulse-level instructions. Implementation at the gate-level allows us to identify
    measurements without wondering about the intent of pulses on the measure channel or
    acquisitions (which might be used for other reasons).
    """

    def run(self, ir: QatIR, *args, **kwargs):
        raise NotImplementedError(
            "The mid-circuit measurement validator needs to be implemented."
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from abc import ABC, abstractmethod

import numpy as np
from pydantic import NonNegativeInt

from qat.ir.gates.base import Angle, GateBase


class Gate1Q(GateBase, ABC):
    """Base gate for single qubit operations."""

    qubit: NonNegativeInt

    @property
    @abstractmethod
    def matrix(self):
        pass

    @property
    def qubits(self):
        return (self.qubit,)


class Id(Gate1Q):
    r"""The identity gate for a single qubit.

    Matrix representation:

    .. math:: I = \begin{bmatrix}1 & 0 \\ 0 & 1 \end{bmatrix}.

    :param qubit: Target qubit index.
    """

    @property
    def matrix(self):
        return np.array([[1, 0], [0, 1]])


class X(Gate1Q):
    r"""Implements the Pauli-X gate.

    Matrix representation:

    .. math:: X = \begin{bmatrix}0 & 1 \\ 1 & 0 \end{bmatrix}.

    :param qubit: Target qubit index.

    """

    @property
    def matrix(self):
        return np.array([[0, 1], [1, 0]])


class Y(Gate1Q):
    r"""Implements the Pauli-Y gate.

    Matrix representation:

    .. math:: Y = \begin{bmatrix} 0 & -i \\ i & 0 \end{bmatrix}.

    :param qubit: Target qubit index.
    """

    @property
    def matrix(self):
        return np.array([[0, -1j], [1j, 0]])


class Z(Gate1Q):
    r"""Implements the Pauli-Z gate.

    Matrix representation:

    .. math:: X = \begin{bmatrix}0 & 1 \\ 1 & 0 \end{bmatrix}.

    :param qubit: Target qubit index.
    """

    @property
    def matrix(self):
        return np.array([[1, 0], [0, -1]])


class Rx(Gate1Q):
    r"""The :math:`R_{x}(\theta) = e^{-i\theta X/2}` gate for some rotation angle
    :math:`\theta`.

    Matrix representation:

    .. math:: R_{x}(\theta) = \begin{bmatrix}
        \cos(\theta/2) & -i\sin(\theta/2) \\ 
        -i\sin(\theta/2) & \cos(\theta/2)
        \end{bmatrix}.

    :param qubit: Target qubit index.
    :param theta: Rotation angle
    """

    theta: Angle = np.pi

    @property
    def matrix(self):
        cos = np.cos(self.theta / 2)
        sin = np.sin(self.theta / 2)
        return np.array([[cos, -1j * sin], [-1j * sin, cos]])


class Ry(Gate1Q):
    r"""The :math:`R_{y}(\theta) = e^{-i\theta Y/2}` gate for some rotation angle
    :math:`\theta`.

    Matrix representation:

    .. math:: R_{y}(\theta) = \begin{bmatrix}
        \cos(\theta/2) & -\sin(\theta/2) \\ 
        \sin(\theta/2) & \cos(\theta/2)
        \end{bmatrix}.

    :param qubit: Target qubit index.
    :param theta: Rotation angle
    """

    theta: Angle = np.pi

    @property
    def matrix(self):
        cos = np.cos(self.theta / 2)
        sin = np.sin(self.theta / 2)
        return np.array([[cos, -sin], [sin, cos]])


class Rz(Gate1Q):
    r"""The :math:`R_{z}(\theta) = e^{-i\theta Z/2}` gate for some rotation angle
    :math:`\theta`.

    Matrix representation:

    .. math:: R_{z}(\theta) = \begin{bmatrix}
        e^{-i\theta/2} & 0 \\ 
        0 & e^{i\theta/2}
        \end{bmatrix}.

    :param qubit: Target qubit index.
    :param theta: Rotation angle
    """

    theta: Angle = np.pi

    @property
    def matrix(self):
        return np.array(
            [[np.exp(-1j * self.theta / 2), 0], [0, np.exp(1j * self.theta / 2)]]
        )


class Phase(Gate1Q):
    r"""The :math:`{\rm Phase}(\theta)` gate for some rotation angle :math:`\theta`.

    Equivalent to the :math:`Rz(\theta)` gate up to a global rotation. Matrix
    representation:

    .. math:: {\rm Phase}(\theta) = \begin{bmatrix}  1 & 0 \\ 0 & e^{i\theta}\end{bmatrix}.

    :param qubit: Target qubit index.
    :param theta: Rotation angle
    """

    theta: Angle = np.pi

    @property
    def matrix(self):
        return np.array([[1, 0], [0, np.exp(1j * self.theta)]])


class U(Gate1Q):
    r"""Implements the :math:`U(\theta, \phi, \lambda)` rotation gate defined by three
    rotation angles.

    The gate can be expressed as

    .. math::
        U(\theta, \phi, \lambda) = 
        e^{\frac{1}{2}i\pi(\lambda+\phi)}R_{z}(\phi)R_{y}(\theta)R_{z}(\lambda)

    Matrix representation:

    .. math::
        U(\theta, \phi, \lambda) = \begin{bmatrix}
        \cos(\theta/2) & -e^{i\lambda} \sin(\theta/2) \\
        e^{i\phi} \sin(\theta/2) & e^{i(\phi+\lambda)} \cos(\theta/2)
        \end{bmatrix}.

    
    :param qubit: Target qubit index.
    :param theta: Rotation angle :math:`\theta` around the Y-axis.
    :param phi: Rotation angle :math:`\phi` around the Z-axis (final).
    :param lambd: Rotation angle :math:`\lambda` around the Z-axis (first).
    """

    theta: Angle = 0.0
    phi: Angle = 0.0
    lambd: Angle = 0.0

    @property
    def matrix(self):
        return np.array(
            [
                [np.cos(self.theta / 2), -np.exp(1j * self.lambd) * np.sin(self.theta / 2)],
                [
                    np.exp(1j * self.phi) * np.sin(self.theta / 2),
                    np.exp(1j * (self.phi + self.lambd)) * np.cos(self.theta / 2),
                ],
            ]
        )

    @classmethod
    def from_matrix(cls, qubit: int, mat: np.ndarray):
        r"""Returns a :math:`U(\theta, \phi, \lambda)` gate directly from a 1Q matrix.

        Determines the angles using simple algebra.
        """
        if np.isclose(mat[0, 0], 0.0):
            theta = np.pi
            lambd = np.angle(-mat[0, 1])
            phi = np.angle(mat[1, 0])
        elif np.isclose(mat[1, 0], 0.0):
            theta = 0.0
            lambd = 0.0
            phi = np.angle(mat[1, 1]) - np.angle(mat[0, 0])
        else:
            alpha = np.angle(mat[0, 0])
            lambd = np.angle(-mat[0, 1]) - alpha
            phi = np.angle(mat[1, 0]) - alpha
            theta = 2 * np.arccos(np.abs(mat[0, 0]))

        return cls(qubit=qubit, theta=theta, phi=phi, lambd=lambd)


class Hadamard(Gate1Q):
    r"""Implements the Hadamard gate :math:`H`.

    Matrix representation:

    .. math::
        H = \begin{bmatrix} 1 & 1 \\ 1 & -1 \end{bmatrix}.
    """

    @property
    def matrix(self):
        return 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])


class S(Gate1Q):
    r"""Implements the :math:`S` gate.

    This gate is identical to the :math:`R_{z}(\pi/2)` gate up to global phase. Matrix
    representation:

    .. math:: S = \begin{bmatrix} 1 & 0 \\ 0 & i \end{bmatrix}.
    """

    @property
    def matrix(self):
        return np.array([[1, 0], [0, 1j]])


class Sdg(Gate1Q):
    r"""Implements the adjoint of the :math:`S` gate, :math:`S^{\dagger}`.

    This gate is identical to the :math:`R_{z}(-\pi/2)` gate up to global phase. Matrix
    representation:

    .. math:: S^{\dagger} = \begin{bmatrix} 1 & 0 \\ 0 & -i \end{bmatrix}.
    """

    @property
    def matrix(self):
        return np.array([[1, 0], [0, -1j]])


class T(Gate1Q):
    r"""Implements the :math:`T` gate.

    This gate is identical to the :math:`R_{z}(\pi/4)` gate up to global phase. Matrix
    representation:

    .. math:: T = \begin{bmatrix} 1 & 0 \\ 0 & (1 + i)/\sqrt{2} \end{bmatrix}.
    """

    @property
    def matrix(self):
        return np.array([[1, 0], [0, (1 + 1j) / (np.sqrt(2))]])


class Tdg(Gate1Q):
    r"""Implements the adjoint of the :math:`T` gate, :math:`T^{\dagger}`.

    This gate is identical to the :math:`R_{z}(-\pi/4)` gate up to global phase. Matrix
    representation:

    .. math:: T^{\dagger} = \begin{bmatrix} 1 & 0 \\ 0 & (1 - i)/\sqrt{2} \end{bmatrix}."""

    @property
    def matrix(self):
        return np.array([[1, 0], [0, (1 - 1j) / (np.sqrt(2))]])



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
import uuid
from abc import ABC
from typing import TypeVar

from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class ResultConcept(ABC):
    """Base class describing the abstraction of an analysis result.

    See :class:`ResultManager`.
    """

    pass


class ResultModel(ResultConcept):
    """Wrapper for any result object typically produced by an analysis pass.

    See :class:`ResultManager`.
    """

    def __init__(self, res_obj):
        self._result = res_obj
        self._uuid = uuid.uuid4()

        self.is_valid = True

    @property
    def value(self):
        return self._result

    @property
    def id(self):
        return self._uuid

    def __hash__(self):
        return hash(self._uuid)


class ResultInfoMixin(ABC):
    """Base mixin specifying result identification mechanism. A result has an :attr:`id`,
    :attr:`name`, and :attr:`value`.

    See :class:`ResultManager`.
    """

    def id(self):
        pass

    def name(self):
        pass

    def value(self):
        pass


ResultType = TypeVar("ResultType", bound=ResultInfoMixin)


class ResultManager:
    """Represents a collection of analysis results with caching and aggregation
    capabilities.

    Passes that merely compute analyses on the IR must not invalidate prior results. Passes
    that mutate any IR units are likely to invalidate predecessor results.

    An analysis pass can produce 1 or more result objects. There is in theory a duality
    between passes and the results they produce.

    To keep things simple, the ResultManager is just a set of analysis results. Result
    identification is also kept trivial where a UUID is used internally. Proper
    identification mechanism will be called upon once we feel the need for a more
    sophisticated :class:`PassManager`.

    See :class:`PassManager`.
    """

    def __init__(self):
        self._results: set[ResultModel] = set()

    @property
    def results(self):
        return self._results

    def _find(self, res_obj):
        found = [res for res in self._results if res.value == res_obj]
        if not found:
            raise ValueError(f"Could not find result {res_obj}")

        if len(found) > 1:
            raise ValueError("Found multiple results")
        return found[0]

    def _remove(self, *res_objs):
        for res_obj in res_objs:
            found = self._find(res_obj)
            self._results.remove(found)

    def cleanup(self):
        self._results = set(res for res in self._results if res.is_valid)

    def update(self, other_res_mgr):
        """Add the results from another results manager.

        :param ResultManager other_res_mgr:
        """
        if not isinstance(other_res_mgr, ResultManager):
            raise ValueError(
                f"Invalid type, expected {ResultManager}, but got {type(other_res_mgr)}"
            )
        self._results.update(other_res_mgr._results)

    def add(self, res_obj: ResultInfoMixin):
        """Add a results object to the manager, overwriting any previous results of the same
        type.

        :param res_obj: Results from a pass, typically an analysis pass.
        """
        found = [res for res in self._results if isinstance(res.value, type(res_obj))]
        self._results -= set(found)
        self._results.add(ResultModel(res_obj))

    def mark_as_dirty(self, *res_objs):
        results = [self._find(res_obj) for res_obj in res_objs]
        for result in results:
            result.is_valid = False

    def lookup_by_type(self, ty: type[ResultType]) -> ResultType:
        """Find a result by its type. Throws an error if none or multiple are found.

        :param ty: The results type.
        """
        found = [res.value for res in self._results if isinstance(res.value, ty)]
        if not found:
            raise ValueError(f"Could not find any results instances of {ty}")

        if len(found) > 1:
            raise ValueError(f"Found multiple results instances of {ty}")

        return found[0]

    def remove_by_type(self, ty: type[ResultType]):
        found = [res for res in self._results if isinstance(res.value, ty)]
        if not found:
            raise ValueError(f"Could not find any results instances of {ty}")

        if len(found) > 1:
            raise ValueError(f"Found multiple results instances of {ty}")

        self._results.remove(found[0])

    def check_for_type(self, ty: type) -> bool:
        """Checks the result manager contains a result with a given type.

        :param ty: The results type.
        """

        found = [res.value for res in self._results if isinstance(res.value, ty)]
        return len(found) > 0


class PreservedResults:
    """
    A mechanism for result invalidation and preservation. Similar to LLVM's new PassManager, we state
    that each transform pass must declare what analysis results it preserves. In this case, the pass returns
    a PreservedResults instance which is then used by the :class:`PassManager` for cache housekeeping.
    """

    @staticmethod
    def all():
        """
        Indicates that **all** of the previously computed results remain valid. The pass manager will do nothing
        as analyses results are still safe and correct after the pass in question has finished running.
        """

        # TODO - Implement PreservedResults [COMPILER-843]
        pass

    @staticmethod
    def none():
        """
        Indicates that **none** of the previously computed results are valid. This will cause the pass manager
        to evict cached results after the (transform) pass in question has finished running.
        """

        # TODO - Implement PreservedResults [COMPILER-843]
        pass

    @staticmethod
    def preserve(*res_obj):
        """
        Selective preservation of the results or indeed result sets indicated by the argument. This will
        cause the pass manager to filter through the result cache and preserve **only** the ones that
        the argument points to.
        """

        # TODO - Implement PreservedResults [COMPILER-843]
        pass

    @staticmethod
    def discard(*res_obj):
        """
        Convenient inverse selector for the :meth:`preserve()` API. This will cause the pass manager
        to discard and evict **only** results pointed to by the argument.
        """

        # TODO - Implement PreservedResults [COMPILER-843]
        pass



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from pathlib import Path

from qat.model.loaders.purr.base import BaseLegacyModelLoader
from qat.purr.backends.qblox.live import QbloxLiveHardwareModel
from qat.purr.compiler.hardware_models import (
    QuantumHardwareModel as LegacyQuantumHardwareModel,
)
from qat.utils.hardware_model import hash_calibration_file


class FileModelLoader(BaseLegacyModelLoader):
    def __init__(self, path: Path | str):
        self.path: Path = Path(path).expanduser()

    def load(self) -> LegacyQuantumHardwareModel:
        hw = LegacyQuantumHardwareModel.load_calibration_from_file(str(self.path))
        hw.calibration_id = hash_calibration_file(str(self.path))
        return hw


class QbloxFileModelLoader(FileModelLoader):
    """
    File loader for the open source QbloxLiveHardwareModel
    """

    def load(self) -> QbloxLiveHardwareModel:
        model = QbloxLiveHardwareModel().load_calibration_from_file(self.path)
        model.calibration_id = hash_calibration_file(self.path)
        model._reverse_coercion()  ## Required to revert strings into integers
        return model



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
from __future__ import annotations

import re
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, TypeVar, Union

import numpy as np

from qat.purr.compiler.devices import (
    Calibratable,
    ChannelType,
    PhysicalBaseband,
    PhysicalChannel,
    PulseChannel,
    QuantumDevice,
    Qubit,
    QubitCoupling,
    Resonator,
)
from qat.purr.compiler.instructions import (
    Acquire,
    AcquireMode,
    CrossResonanceCancelPulse,
    CrossResonancePulse,
    DrivePulse,
    Instruction,
    PhaseShift,
    Synchronize,
)
from qat.purr.utils.logger import get_default_logger
from qat.purr.utils.logging_utils import log_duration

log = get_default_logger()

if TYPE_CHECKING:
    from qat.purr.compiler.builders import InstructionBuilder, QuantumInstructionBuilder
    from qat.purr.compiler.execution import InstructionExecutionEngine

AnyEngine = TypeVar("AnyEngine")
AnyBuilder = TypeVar("AnyBuilder")


class HardwareModel:
    """
    Base class for all hardware models. Every model should return the builder class that
    should be used to build circuits/pulses for its particular back-end.
    """

    def __init__(self, shot_limit=-1):
        super().__init__()
        self.repeat_limit = shot_limit

    def create_engine(self) -> InstructionExecutionEngine: ...

    def create_runtime(self, existing_engine: InstructionExecutionEngine = None):
        if existing_engine is None:
            existing_engine = self.create_engine()

        from qat.purr.compiler.runtime import QuantumRuntime

        return QuantumRuntime(existing_engine)

    def create_builder(self) -> InstructionBuilder: ...

    def __repr__(self):
        return self.__class__.__name__


def get_cl2qu_index_mapping(instructions: List[Instruction], model: QuantumHardwareModel):
    """
    Returns a Dict[str, str] mapping creg to qreg indices.
    Classical register indices are extracted following the pattern <clreg_name>[<clreg_index>]
    """
    mapping = {}
    pattern = re.compile(r"(.*)\[([0-9]+)\]")

    for instruction in instructions:
        if not isinstance(instruction, Acquire):
            continue

        qubit = next(
            (
                qubit
                for qubit in model.qubits
                if qubit.get_acquire_channel().full_id() == instruction.channel.full_id()
            ),
            None,
        )
        if qubit is None:
            raise ValueError(
                f"Could not find any qubits by acquire channel {instruction.channel}"
            )

        result = pattern.match(instruction.output_variable)
        if result is None:
            raise ValueError(
                f"Could not extract cl register index from {instruction.output_variable}"
            )

        clbit_index = result.group(2)
        mapping[clbit_index] = qubit.index

    return mapping


class ReadoutMitigation(Calibratable):
    """
    Linear maps each individual qubit to its <0/1> given <0/1> probability.
    Note that linear assumes no cross-talk effects and considers each qubit independent.
    linear = {
        <qubit_number>: {
            "0|0": 1,
            "1|0": 1,
            "0|1": 1,
            "1|1": 1,
        }
    }
    Matrix is the entire 2**n x 2**n process matrix of p(<bitstring_1>|<bitstring_2>).
    M3 is a runtime mitigation strategy that builds up the calibration it needs at runtime,
    hence a bool of available or not. For more info https://github.com/Qiskit-Partners/mthree.
    """

    def __init__(
        self,
        linear: [Dict[str, Dict[str, float]]] = None,
        matrix: np.array = None,
        m3: bool = False,
    ):
        super().__init__()
        self.m3_available = m3
        self.linear: [Dict[str, Dict[str, float]]] = linear
        self.matrix: np.array = matrix


class ErrorMitigation(Calibratable):
    def __init__(self, readout_mitigation: Optional[ReadoutMitigation] = None):
        super().__init__()
        self.readout_mitigation = readout_mitigation

    def __setstate__(self, state):
        readout_mitigation = state.pop("readout_mitigation", None)
        if readout_mitigation is None or isinstance(readout_mitigation, ReadoutMitigation):
            self.readout_mitigation = readout_mitigation
        else:
            raise ValueError(
                f"Expected {ReadoutMitigation.__class__} but got {type(readout_mitigation)}"
            )

        self.__dict__.update(state)


class QuantumHardwareModel(HardwareModel, Calibratable):
    """
    Object modelling our superconducting hardware. Holds up-to-date information about a
    current piece of hardware, whether simulated or physical machine.
    """

    def __init__(
        self,
        shot_limit=10000,
        acquire_mode=None,
        repeat_count=1000,
        repetition_period=100e-6,
        error_mitigation: Optional[ErrorMitigation] = None,
        calibration_id: str = "",
    ):
        # Our hardware has a default shot limit of 10,000 right now.
        self.default_acquire_mode = acquire_mode or AcquireMode.RAW
        self.default_repeat_count = repeat_count
        self.default_repetition_period = repetition_period

        self.quantum_devices: Dict[str, QuantumDevice] = {}
        self.pulse_channels: Dict[str, PulseChannel] = {}
        self.physical_channels: Dict[str, PhysicalChannel] = {}
        self.basebands: Dict[str, PhysicalBaseband] = {}
        self.qubit_direction_couplings: List[QubitCoupling] = []
        self.error_mitigation: Optional[ErrorMitigation] = error_mitigation
        self.calibration_id = calibration_id

        # Construct last due to us overriding calibratables fields with properties.
        super().__init__(
            shot_limit=shot_limit,
        )

    def qubit_quality(self, qubit_index: int):
        """
        Readout quality for a qubit with index `qubit_index`.
        Note that this does not take gate fidelity into account.
        """
        readout_quality = self.error_mitigation.readout_mitigation.linear.get(
            str(qubit_index), None
        )

        if readout_quality:
            return (readout_quality["0|0"] + readout_quality["1|1"]) / 2
        else:
            log.warning(f"Qubit with index {qubit_index} not found.")
            return 0

    def create_engine(self) -> InstructionExecutionEngine:
        from qat.purr.backends.echo import EchoEngine

        return EchoEngine(self)

    def create_builder(self) -> "QuantumInstructionBuilder":
        from qat.purr.compiler.builders import QuantumInstructionBuilder

        return QuantumInstructionBuilder(self)

    def resolve_qb_pulse_channel(
        self, chanbit: Union[Qubit, PulseChannel]
    ) -> Tuple[Qubit, PulseChannel]:
        if isinstance(chanbit, Qubit):
            return chanbit, chanbit.get_default_pulse_channel()
        else:
            for qubit in self.qubits:
                for channel in qubit.pulse_channels.values():
                    if chanbit == channel:
                        return qubit, chanbit
        raise ValueError(f"Cannot resolve {chanbit}")

    @property
    def qubits(self):
        """Returns list of the qubits on this hardware sorted by index."""
        qubits = [qb for qb in self.quantum_devices.values() if isinstance(qb, Qubit)]
        qubits = sorted(qubits, key=lambda x: x.index)
        return qubits

    @property
    def resonators(self):
        return [qd for qd in self.quantum_devices.values() if isinstance(qd, Resonator)]

    def has_qubit(self, id_: Union[int, str]):
        if isinstance(id_, int):
            id_ = f"Q{id_}"
        return id_ in self.quantum_devices

    def get_qubit(self, id_: Union[int, str, Qubit]) -> Qubit:
        """
        Returns a qubit based on id/index. If the passed-in object is already a Qubit
        object just returns that.
        """
        if isinstance(id_, Qubit):
            return id_

        if isinstance(id_, int):
            id_ = f"Q{id_}"

        if id_ not in self.quantum_devices:
            raise ValueError(f"Tried to retrieve a qubit ({str(id_)}) that doesn't exist.")

        found_qubit = self.quantum_devices.get(id_)
        if not isinstance(found_qubit, Qubit):
            raise ValueError(f"Device with id {str(id_)} isn't a qubit.")

        return found_qubit

    def add_quantum_device(self, *devices: QuantumDevice):
        for device in devices:
            existing_dev = self.quantum_devices.get(device.full_id(), None)
            if existing_dev is not None:
                # If we're the same instance just don't throw.
                if existing_dev is device:
                    continue

                raise KeyError(f"Quantum Device with id '{device.id}' already exists.")

            self.quantum_devices[device.full_id()] = device

            # add device pulse channels to hardware if not already present
            for pulse_channel in device.pulse_channels.values():
                self.add_pulse_channel(pulse_channel)

    def get_quantum_device(self, id_: str):
        return self.quantum_devices.get(id_, None)

    def add_pulse_channel(self, *pulse_channels: PulseChannel):
        for pulse_channel in pulse_channels:
            existing_channel = self.pulse_channels.get(pulse_channel.full_id(), None)
            if existing_channel is not None:
                # If we're the same instance just don't throw.
                if existing_channel is pulse_channel:
                    continue

                raise KeyError(
                    f"Pulse channel with id '{pulse_channel.full_id()}' already exists."
                )

            self.pulse_channels[pulse_channel.full_id()] = pulse_channel

        return pulse_channels

    def get_pulse_channel_from_device(
        self,
        ch_type: ChannelType,
        host_device_id: str,
        aux_device_ids: List[str] = None,
    ):
        if aux_device_ids is None:
            aux_device_ids = []
        return self.get_quantum_device(host_device_id).get_pulse_channel(
            ch_type, [self.get_quantum_device(device) for device in aux_device_ids]
        )

    def get_pulse_channels_from_physical_channel(
        self, physical_channel: Union[str, PhysicalChannel]
    ):
        # WARNING: this will not get pulse channels created during execution, only ones
        # predefined on the hardware.
        if isinstance(physical_channel, str):
            physical_channel = self.get_physical_channel(physical_channel)
        pulse_channels = []
        for pulse_channel in self.pulse_channels.values():
            if pulse_channel.physical_channel == physical_channel:
                pulse_channels.append(pulse_channel)

        return pulse_channels

    def get_devices_from_pulse_channel(self, pulse_channel: Union[str, PulseChannel]):
        if isinstance(pulse_channel, str):
            pulse_channel = self.get_pulse_channel_from_id(pulse_channel)
        devices = [
            device
            for device in self.quantum_devices.values()
            if pulse_channel in device.pulse_channels.values()
        ]
        return devices

    def get_pulse_channel_from_id(self, id_: str):
        return self.pulse_channels.get(id_)

    def get_devices_from_physical_channel(
        self, physical_channel: Union[str, PhysicalChannel]
    ):
        if isinstance(physical_channel, str):
            physical_channel = self.get_physical_channel(physical_channel)
        devices = [
            device
            for device in self.quantum_devices.values()
            if physical_channel == device.physical_channel
        ]
        return devices

    def add_physical_channel(self, *physical_channels: PhysicalChannel):
        for physical_channel in physical_channels:
            existing_channel = self.physical_channels.get(physical_channel.full_id(), None)
            if existing_channel is not None:
                # If we're the same instance just don't throw.
                if existing_channel is physical_channel:
                    continue

                raise KeyError(
                    f"Physical channel with id '{physical_channel.full_id()}' already exists."
                )

            self.physical_channels[physical_channel.full_id()] = physical_channel

    def get_physical_channel(self, id_: str):
        return self.physical_channels.get(id_, None)

    def add_physical_baseband(self, *basebands: PhysicalBaseband):
        for baseband in basebands:
            existing_baseband = self.basebands.get(baseband.full_id(), None)
            if existing_baseband is not None:
                # If we're the same instance just don't throw.
                if existing_baseband is baseband:
                    continue

                raise KeyError(f"Baseband with id '{baseband.full_id()}' already exists.")

            self.basebands[baseband.full_id()] = baseband

    def get_physical_baseband(self, id_: str):
        return self.basebands.get(id_, None)

    def get_device(self, id_):
        if (device := self.get_quantum_device(id_)) is not None:
            return device

        if (device := self.get_physical_channel(id_)) is not None:
            return device

        if (device := self.get_pulse_channel_from_id(id_)) is not None:
            return device

        return self.get_physical_baseband(id_)

    def add_device(self, device):
        if isinstance(device, PhysicalBaseband):
            self.add_physical_baseband(device)
        elif isinstance(device, PhysicalChannel):
            self.add_physical_channel(device)
        else:
            self.add_quantum_device(device)

    def _resolve_qb_pulse_channel(self, chanbit: Union[Qubit, PulseChannel]):
        if isinstance(chanbit, Qubit):
            return chanbit, chanbit.get_default_pulse_channel()
        else:
            quantum_devices = self.get_devices_from_pulse_channel(chanbit)
            primary_devices = [
                device
                for device in quantum_devices
                if device.get_default_pulse_channel() == chanbit
            ]
            if len(primary_devices) > 1:
                raise NotImplementedError(
                    f"Too many devices use the channel with id {chanbit.full_id()} as "
                    "a default channel to resolve a primary device."
                )
            elif len(primary_devices) == 0:
                return None, chanbit
            else:
                return primary_devices[0], chanbit

    def _map_resonator_to_qubit(self, resonator: Resonator):
        """Finds and returns the qubit that is coupled to the resonator."""
        if not isinstance(resonator, Resonator):
            raise ValueError(f"Expected Resonator but got {type(resonator)}.")

        for qubit in self.qubits:
            if resonator.full_id() == qubit.measure_device.full_id():
                return qubit

        return None

    @property
    def is_calibrated(self):
        def check_devices(target_devices):
            return all(
                [
                    device.is_calibrated
                    for device in target_devices.values()
                    if isinstance(device, Calibratable)
                ]
            )

        return (
            check_devices(self.quantum_devices)
            and check_devices(self.physical_channels)
            and check_devices(self.basebands)
        )

    @is_calibrated.setter
    def is_calibrated(self, val):
        def set_devices(target_dict):
            for device in target_dict.values():
                if isinstance(device, Calibratable):
                    device.is_calibrated = val

        set_devices(self.quantum_devices)
        set_devices(self.physical_channels)
        set_devices(self.basebands)

    def constrain(self, angle):
        return (angle + np.pi) % (2 * np.pi) - np.pi

    def get_hw_x_pi_2(
        self, qubit, pulse_channel: PulseChannel = None, amp_scale: float = None
    ) -> List[Any]:
        if amp_scale is None or np.isclose(amp_scale, 1.0):
            pulse_args = qubit.pulse_hw_x_pi_2
        else:
            pulse_args = deepcopy(qubit.pulse_hw_x_pi_2)
            pulse_args["amp"] *= amp_scale

        # note: this could be a more complicated set of instructions
        return [
            DrivePulse(pulse_channel or qubit.get_default_pulse_channel(), **pulse_args)
        ]

    def get_hw_x_pi(self, qubit, pulse_channel: PulseChannel = None):
        return [
            DrivePulse(
                pulse_channel or qubit.get_default_pulse_channel(), **qubit.pulse_hw_x_pi
            )
        ]

    def get_hw_z(self, qubit, phase, pulse_channel: PulseChannel = None) -> List[Any]:
        if phase == 0:
            return []

        instr_collection = [
            PhaseShift(pulse_channel or qubit.get_default_pulse_channel(), phase)
        ]
        if pulse_channel is None or pulse_channel == qubit.get_default_pulse_channel():
            instr_collection.extend(
                PhaseShift(coupled_qubit.get_cross_resonance_channel(qubit), phase)
                for coupled_qubit in qubit.coupled_qubits
            )
            # TODO: the cross-resonance cancellation channel mirrors the cr channel,
            #   they should be tied together
            instr_collection.extend(
                PhaseShift(
                    qubit.get_cross_resonance_cancellation_channel(coupled_qubit), phase
                )
                for coupled_qubit in qubit.coupled_qubits
            )

        return instr_collection

    def get_hw_zx_pi_4(self, qubit: Qubit, target_qubit: Qubit) -> List[Any]:
        control_channel = qubit.get_cross_resonance_channel(target_qubit)
        target_channel = target_qubit.get_cross_resonance_cancellation_channel(qubit)
        pulse = qubit.pulse_hw_zx_pi_4.get(target_qubit.id, None)
        if pulse is None:
            raise ValueError(
                f"Tried to perform cross resonance on {str(target_qubit)} "
                f"that isn't linked to {str(qubit)}."
            )

        return [
            Synchronize([control_channel, target_channel]),
            CrossResonancePulse(control_channel, **pulse),
            CrossResonanceCancelPulse(target_channel, **pulse),
        ]

    def get_gate_U(self, qubit, theta, phi, lamb, pulse_channel: PulseChannel = None):
        theta = self.constrain(theta)
        return [
            *self.get_hw_z(qubit, lamb + np.pi, pulse_channel),
            *self.get_hw_x_pi_2(qubit, pulse_channel),
            *self.get_hw_z(qubit, np.pi - theta, pulse_channel),
            *self.get_hw_x_pi_2(qubit, pulse_channel),
            *self.get_hw_z(qubit, phi, pulse_channel),
        ]

    def get_gate_X(self, qubit, theta, pulse_channel: PulseChannel = None):
        """
        Provides pulses to perform an X rotation by angle theta on the given qubit.
        If theta is close to +/- pi/2, the standard x_pi_2 pulse is used, with some phase rotation in the case of - pi.
        If theta is close to +/- pi and the qubit has direct_x_pi set to True,
        the get_hw_x_pi method will return an x pi pulse.
        Otherwise, the gate is decomposed using the U3 decomposition.
        """
        theta = self.constrain(theta)
        if np.isclose(theta, 0.0):
            return []
        elif np.isclose(theta, np.pi / 2.0):
            return self.get_hw_x_pi_2(qubit, pulse_channel)
        elif np.isclose(theta, -np.pi / 2.0):
            return [
                *self.get_hw_z(qubit, np.pi, pulse_channel),
                *self.get_hw_x_pi_2(qubit, pulse_channel),
                *self.get_hw_z(qubit, -np.pi, pulse_channel),
            ]
        elif np.isclose(theta, np.pi) and getattr(qubit, "direct_x_pi", False):
            return self.get_hw_x_pi(qubit, pulse_channel)
        elif np.isclose(theta, -np.pi) and getattr(qubit, "direct_x_pi", False):
            return [
                *self.get_hw_z(qubit, np.pi, pulse_channel),
                *self.get_hw_x_pi(qubit, pulse_channel),
                *self.get_hw_z(qubit, -np.pi, pulse_channel),
            ]
        return self.get_gate_U(qubit, theta, -np.pi / 2.0, np.pi / 2.0, pulse_channel)

    def get_gate_Y(self, qubit, theta, pulse_channel: PulseChannel = None):
        theta = self.constrain(theta)
        if np.isclose(theta, 0.0):
            return []
        elif np.isclose(theta, np.pi / 2.0):
            return [
                *self.get_hw_z(qubit, -np.pi / 2.0, pulse_channel),
                *self.get_hw_x_pi_2(qubit, pulse_channel),
                *self.get_hw_z(qubit, np.pi / 2.0, pulse_channel),
            ]
        elif np.isclose(theta, -np.pi / 2.0):
            return [
                *self.get_hw_z(qubit, np.pi / 2.0, pulse_channel),
                *self.get_hw_x_pi_2(qubit, pulse_channel),
                *self.get_hw_z(qubit, -np.pi / 2.0, pulse_channel),
            ]
        return self.get_gate_U(qubit, theta, 0.0, 0.0, pulse_channel)

    def get_gate_Z(self, qubit, theta, pulse_channel: PulseChannel = None):
        theta = self.constrain(theta)
        return self.get_hw_z(qubit, theta, pulse_channel)

    def get_gate_ZX(self, qubit, theta, target_qubit):
        theta = self.constrain(theta)
        cr_pulse_channel = qubit.get_pulse_channel(
            ChannelType.cross_resonance, [target_qubit]
        )
        cr_cancellation_pulse_channel = target_qubit.get_pulse_channel(
            ChannelType.cross_resonance_cancellation, [qubit]
        )
        if np.isclose(theta, 0.0):
            return []
        elif np.isclose(theta, np.pi / 4.0):
            return self.get_hw_zx_pi_4(qubit, target_qubit)
        elif np.isclose(theta, -np.pi / 4.0):
            return [
                PhaseShift(cr_pulse_channel, np.pi),
                PhaseShift(cr_cancellation_pulse_channel, np.pi),
                *self.get_hw_zx_pi_4(qubit, target_qubit),
                PhaseShift(cr_pulse_channel, np.pi),
                PhaseShift(cr_cancellation_pulse_channel, np.pi),
            ]
        else:
            raise ValueError("Generic ZX gate not implemented yet!")

    def get_calibration(self):
        with log_duration("Saved hardware model to calibration in {}s."):
            return super().get_calibration()

    @classmethod
    def load_calibration(cls, calibration_string: str):
        with log_duration("Loaded hardware model from calibration in {}s."):
            return super().load_calibration(calibration_string)

    def __setstate__(self, state):
        error_mitigation = state.pop("error_mitigation", None)
        if error_mitigation is None or isinstance(error_mitigation, ErrorMitigation):
            self.error_mitigation = error_mitigation
        else:
            raise ValueError(
                f"Expected {ErrorMitigation.__class__} but got {type(error_mitigation)}"
            )

        self.__dict__.update(state)


LegacyHardwareModel = QuantumHardwareModel



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.model.loaders.purr.dummy import QbloxDummyModelLoader
from qat.pipelines.base import BasePipeline
from qat.pipelines.purr.qblox.compile import QbloxCompilePipeline1, QbloxCompilePipeline2
from qat.pipelines.purr.qblox.execute import QbloxExecutePipeline
from qat.pipelines.purr.qblox.full import QbloxPipeline1, QbloxPipeline2
from qat.pipelines.updateable import PipelineConfig


def _create_pipeline_instance(num_qubits: int) -> BasePipeline:
    name = f"dummy{num_qubits}"
    return QbloxPipeline2(
        config=PipelineConfig(name=name),
        loader=QbloxDummyModelLoader(name=name, qubit_count=num_qubits),
    ).pipeline


dummy8 = _create_pipeline_instance(8)
dummy16 = _create_pipeline_instance(16)
# dummy32 = _create_pipeline_instance(32) # TODO: 32Q support: COMPILER-728


__all__ = [
    "QbloxPipeline1",
    "QbloxPipeline2",
    "QbloxExecutePipeline",
    "dummy8",
    "dummy16",
    # "dummy32",  # TODO: 32Q support: COMPILER-728
    "PipelineConfig",
    "QbloxCompilePipeline1",
    "QbloxCompilePipeline2",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from warnings import warn

from qat.engines.waveform.echo import EchoEngine as NewEchoEngine


class EchoEngine(NewEchoEngine):
    """The :class:`EchoEngine` is a minimal execution engine primarily used for testing the
    compilation pipeline.

    It is not connected to any target machine such as live hardware or a simulator, and just
    simply "echos" back the buffers. It only accepts :class:`WaveformProgram` programs.

    It checks the shape of the results generated match the expected shapes from codegen,
    to test consistency across the compilation and execution pipeline.
    """

    def __init__(self, *args, **kwargs):
        warn(
            "WaveformV1 support is deprecated and will be removed in v4.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)
